#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on April 2020
@authors:
Thomas Bonald <bonald@enst.fr>
Quentin Lutz <qlutz@enst.fr>
"""
from typing import Optional, Iterable, Union, Tuple

import numpy as np
from scipy import sparse

from sknetwork.clustering.louvain import BiLouvain
from sknetwork.embedding.spring import Spring
from sknetwork.visualization.colors import STANDARD_COLORS, COOLWARM_RGB


def min_max_scaling(x: np.ndarray, x_min: Optional[float] = None, x_max: Optional[float] = None) -> np.ndarray:
    """Shift and scale vector to be between 0 and 1."""
    x = x.astype(float)
    if x_min is None:
        x_min = np.min(x)
    if x_max is None:
        x_max = np.max(x)
    x -= x_min
    if x_max > x_min:
        x /= (x_max - x_min)
    else:
        x = .5 * np.ones_like(x)
    return x


def rescale(position: np.ndarray, width: float, height: float, margin: float, node_size: float, node_size_max: float,
            display_node_weight: bool):
    """Rescale position and adjust parameters.

    Parameters
    ----------
    position :
        array to rescale
    width :
        Horizontal scaling parameter
    height :
        Vertical scaling parameter
    margin :
        Minimal margin for the plot
    node_size :
        Node size (used to adapt the margin)
    node_size_max :
        Maximum node size (used to adapt the margin)
    display_node_weight :
        If ``True``, display node weight (used to adapt the margin)

    Returns
    -------
    position :
        Rescaled positions
    width :
        Rescaled width
    height :
        Rescaled height
    """
    x = position[:, 0]
    y = position[:, 1]
    span_x = np.max(x) - np.min(x)
    span_y = np.max(y) - np.min(y)
    x = min_max_scaling(x)
    y = 1 - min_max_scaling(y)
    position = np.vstack((x, y)).T

    # rescale
    if width and not height:
        height = width
        if span_x and span_y:
            height *= span_y / span_x
    elif height and not width:
        width = height
        if span_x and span_y:
            width *= span_x / span_y

    position = position * np.array([width, height])

    # margins
    margin = max(margin, node_size_max * display_node_weight, node_size)
    position += margin
    width += 2 * margin
    height += 2 * margin
    return position, width, height


def get_label_colors(label_colors: Optional[Iterable]):
    """Return label svg colors.

    Examples
    --------
    >>> get_label_colors(['black'])
    array(['black'], dtype='<U5')
    >>> get_label_colors({0: 'blue'})
    array(['blue'], dtype='<U64')
    """
    if label_colors is not None:
        if isinstance(label_colors, dict):
            keys = list(label_colors.keys())
            values = list(label_colors.values())
            label_colors = np.array(['black'] * (max(keys) + 1), dtype='U64')
            label_colors[keys] = values
        elif isinstance(label_colors, list):
            label_colors = np.array(label_colors)
    else:
        label_colors = STANDARD_COLORS.copy()
    return label_colors


def get_node_colors(n: int, labels: Optional[Iterable], scores: Optional[Iterable],
                    membership: Optional[sparse.csr_matrix],
                    node_color: str, label_colors: Optional[Iterable],
                    score_min: Optional[float] = None, score_max: Optional[float] = None) -> np.ndarray:
    """Return the colors of the nodes using either labels or scores or default color."""
    node_colors = np.array(n * [node_color]).astype('U64')
    if labels is not None:
        if isinstance(labels, dict):
            keys = np.array(list(labels.keys()))
            values = np.array(list(labels.values())).astype(int)
            labels = -np.ones(n, dtype=int)
            labels[keys] = values
        elif isinstance(labels, list):
            if len(labels) != n:
                raise ValueError("The number of labels must be equal to the corresponding number of nodes.")
            else:
                labels = np.array(labels)
        index = labels >= 0
        label_colors = get_label_colors(label_colors)
        node_colors[index] = label_colors[labels[index] % len(label_colors)]
    elif scores is not None:
        colors_score = COOLWARM_RGB.copy()
        n_colors = colors_score.shape[0]
        colors_score_svg = np.array(['rgb' + str(tuple(colors_score[i])) for i in range(n_colors)])
        if isinstance(scores, dict):
            keys = np.array(list(scores.keys()))
            values = np.array(list(scores.values()))
            scores = (min_max_scaling(values, score_min, score_max) * (n_colors - 1)).astype(int)
            node_colors[keys] = colors_score_svg[scores]
        else:
            if isinstance(scores, list):
                if len(scores) != n:
                    raise ValueError("The number of scores must be equal to the corresponding number of nodes.")
                else:
                    scores = np.array(scores)
            scores = (min_max_scaling(scores, score_min, score_max) * (n_colors - 1)).astype(int)
            node_colors = colors_score_svg[scores]
    elif membership is not None:
        if isinstance(label_colors, dict):
            raise TypeError("Label colors must be a list or an array when using a membership.")
        label_colors = get_label_colors(label_colors)
        node_colors = label_colors
    return node_colors


def get_node_widths(n: int, seeds: Union[int, dict, list], node_width: float, node_width_max: float) -> np.ndarray:
    """Return the node widths."""
    node_widths = node_width * np.ones(n)
    if seeds is not None:
        if type(seeds) == dict:
            seeds = list(seeds.keys())
        elif np.issubdtype(type(seeds), np.integer):
            seeds = [seeds]
        if len(seeds):
            node_widths[np.array(seeds)] = node_width_max
    return node_widths


def get_node_sizes(weights: np.ndarray, node_size: float, node_size_min: float, node_size_max: float, node_weight) \
        -> np.ndarray:
    """Return the node sizes."""
    if node_weight and np.min(weights) < np.max(weights):
        node_sizes = node_size_min + np.abs(node_size_max - node_size_min) * weights / np.max(weights)
    else:
        node_sizes = node_size * np.ones_like(weights)
    return node_sizes


def get_node_sizes_bipartite(weights_row: np.ndarray, weights_col: np.ndarray, node_size: float, node_size_min: float,
                             node_size_max: float, node_weight) -> (np.ndarray, np.ndarray):
    """Return the node sizes for bipartite graphs."""
    weights = np.hstack((weights_row, weights_col))
    if node_weight and np.min(weights) < np.max(weights):
        node_sizes_row = node_size_min + np.abs(node_size_max - node_size_min) * weights_row / np.max(weights)
        node_sizes_col = node_size_min + np.abs(node_size_max - node_size_min) * weights_col / np.max(weights)
    else:
        node_sizes_row = node_size * np.ones_like(weights_row)
        node_sizes_col = node_size * np.ones_like(weights_col)
    return node_sizes_row, node_sizes_col


def get_edge_colors(adjacency: sparse.csr_matrix, edge_labels: Optional[list], edge_color: str,
                    label_colors: Optional[Iterable]) -> Tuple[np.ndarray, np.ndarray, list]:
    """Return the edge colors."""
    n_row, n_col = adjacency.shape
    n_edges = adjacency.nnz
    adjacency_labels = (adjacency > 0).astype(int)
    adjacency_labels.data = -adjacency_labels.data
    edge_colors_residual = []
    if edge_labels:
        label_colors = get_label_colors(label_colors)
        for i, j, label in edge_labels:
            if i < 0 or i >= n_row or j < 0 or j >= n_col:
                raise ValueError('Invalid node index in edge labels.')
            if adjacency[i, j]:
                adjacency_labels[i, j] = label % len(label_colors)
            else:
                color = label_colors[label % len(label_colors)]
                edge_colors_residual.append((i, j, color))
    edge_order = np.argsort(adjacency_labels.data)
    edge_colors = np.array(n_edges * [edge_color]).astype('U64')
    index = np.argwhere(adjacency_labels.data >= 0).ravel()
    if len(index):
        edge_colors[index] = label_colors[adjacency_labels.data[index]]
    return edge_colors, edge_order, edge_colors_residual


def get_edge_widths(adjacency: sparse.coo_matrix, edge_width: float, edge_width_min: float, edge_width_max: float,
                    display_edge_weight: bool) -> np.ndarray:
    """Return the edge widths."""
    weights = adjacency.data
    edge_widths = None
    if len(weights):
        if display_edge_weight and np.min(weights) < np.max(weights):
            edge_widths = edge_width_min + np.abs(edge_width_max - edge_width_min) * (weights - np.min(weights))\
                          / (np.max(weights) - np.min(weights))
        else:
            edge_widths = edge_width * np.ones_like(weights)
    return edge_widths


def svg_node(pos_node: np.ndarray, size: float, color: str, stroke_width: float = 1, stroke_color: str = 'black') \
        -> str:
    """Return svg code for a node.

    Parameters
    ----------
    pos_node :
        (x, y) coordinates of the node.
    size :
        Radius of disk in pixels.
    color :
        Color of the disk in SVG valid format.
    stroke_width :
        Width of the contour of the disk in pixels, centered around the circle.
    stroke_color :
        Color of the contour in SVG valid format.

    Returns
    -------
    SVG code for the node.
    """
    x, y = pos_node.astype(int)
    return """<circle cx="{}" cy="{}" r="{}" style="fill:{};stroke:{};stroke-width:{}"/>\n"""\
        .format(x, y, size, color, stroke_color, stroke_width)


def svg_pie_chart_node(pos_node: np.ndarray, size: float, membership: np.ndarray, colors: np.ndarray,
                       stroke_width: float = 1, stroke_color: str = 'black') -> str:
    """Return svg code for a pie-chart node."""
    x, y = pos_node.astype(float)
    n_colors = len(colors)
    out = ""
    cumsum = np.zeros(membership.shape[1] + 1)
    cumsum[1:] = np.cumsum(membership)
    if cumsum[-1] == 0:
        return svg_node(pos_node, size, 'white', stroke_width=3)
    sum_membership = cumsum[-1]
    cumsum = np.multiply(cumsum, (2 * np.pi) / cumsum[-1])
    x_array = size * np.cos(cumsum) + x
    y_array = size * np.sin(cumsum) + y
    large = np.array(membership > sum_membership / 2).ravel()
    for index in range(membership.shape[1]):
        out += """<path d="M {} {} A {} {} 0 {} 1 {} {} L {} {}" style="fill:{};stroke:{};stroke-width:{}" />\n"""\
            .format(x_array[index], y_array[index], size, size, int(large[index]),
                    x_array[index + 1], y_array[index + 1], x, y, colors[index % n_colors], stroke_color, stroke_width)
    return out


def svg_edge(pos_1: np.ndarray, pos_2: np.ndarray, edge_width: float = 1, edge_color: str = 'black') -> str:
    """Return svg code for an edge."""
    x1, y1 = pos_1.astype(int)
    x2, y2 = pos_2.astype(int)
    return """<path stroke-width="{}" stroke="{}" d="M {} {} {} {}"/>\n"""\
        .format(edge_width, edge_color, x1, y1, x2, y2)


def svg_edge_directed(pos_1: np.ndarray, pos_2: np.ndarray, edge_width: float = 1, edge_color: str = 'black',
                      node_size: float = 1.) -> str:
    """Return svg code for a directed edge."""
    vec = pos_2 - pos_1
    norm = np.linalg.norm(vec)
    if norm:
        x, y = ((vec / norm) * node_size).astype(int)
        x1, y1 = pos_1.astype(int)
        x2, y2 = pos_2.astype(int)
        return """<path stroke-width="{}" stroke="{}" d="M {} {} {} {}" marker-end="url(#arrow)"/>\n"""\
            .format(edge_width, edge_color, x1, y1, x2 - x, y2 - y)
    else:
        return ""


def svg_text(pos, text, font_size=12, align_right=False):
    """Return svg code for text."""
    x, y = pos.astype(int)
    text = str(text).replace('&', ' ')
    if align_right:
        return """<text text-anchor="end" x="{}" y="{}" font-size="{}">{}</text>"""\
            .format(x, y, font_size, text)
    else:
        return """<text x="{}" y="{}" font-size="{}">{}</text>"""\
            .format(x, y, font_size, text)


def svg_graph(adjacency: Optional[sparse.csr_matrix] = None, position: Optional[np.ndarray] = None,
              names: Optional[np.ndarray] = None,
              labels: Optional[Iterable] = None, scores: Optional[Iterable] = None,
              membership: Optional[sparse.csr_matrix] = None,
              seeds: Union[list, dict] = None, width: Optional[float] = 400, height: Optional[float] = 300,
              margin: float = 20, margin_text: float = 3, scale: float = 1, node_order: Optional[np.ndarray] = None,
              node_size: float = 7, node_size_min: float = 1, node_size_max: float = 20,
              display_node_weight: bool = False, node_weights: Optional[np.ndarray] = None, node_width: float = 1,
              node_width_max: float = 3, node_color: str = 'gray',
              display_edges: bool = True, edge_labels: Optional[list] = None,
              edge_width: float = 1, edge_width_min: float = 0.5,
              edge_width_max: float = 20, display_edge_weight: bool = True,
              edge_color: Optional[str] = None, label_colors: Optional[Iterable] = None,
              font_size: int = 12, directed: bool = False, filename: Optional[str] = None) -> str:
    """Return SVG image of a graph.

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    position :
        Positions of the nodes.
    names :
        Names of the nodes.
    labels :
        Labels of the nodes (negative values mean no label).
    scores :
        Scores of the nodes (measure of importance).
    membership :
        Membership of the nodes (label distribution).
    seeds :
        Nodes to be highlighted (if dict, only keys are considered).
    width :
        Width of the image.
    height :
        Height of the image.
    margin :
        Margin of the image.
    margin_text :
        Margin between node and text.
    scale :
        Multiplicative factor on the dimensions of the image.
    node_order :
        Order in which nodes are displayed.
    node_size :
        Size of nodes.
    node_size_min :
        Minimum size of a node.
    node_size_max:
        Maximum size of a node.
    node_width :
        Width of node circle.
    node_width_max :
        Maximum width of node circle.
    node_color :
        Default color of nodes (svg color).
    display_node_weight :
        If ``True``, display node weights through node size.
    node_weights :
        Node weights (used only if **display_node_weight** is ``True``).
    display_edges :
        If ``True``, display edges.
    edge_labels :
        Labels of the edges, as a list of tuples (source, destination, label)
    edge_width :
        Width of edges.
    edge_width_min :
        Minimum width of edges.
    edge_width_max :
        Maximum width of edges.
    display_edge_weight :
        If ``True``, display edge weights through edge widths.
    edge_color :
        Default color of edges (svg color).
    label_colors:
        Colors of the labels (svg colors).
    font_size :
        Font size.
    directed :
        If ``True``, considers the graph as directed.
    filename :
        Filename for saving image (optional).

    Returns
    -------
    image : str
        SVG image.

    Example
    -------
    >>> from sknetwork.data import karate_club
    >>> graph = karate_club(True)
    >>> adjacency = graph.adjacency
    >>> position = graph.position
    >>> from sknetwork.visualization import svg_graph
    >>> image = svg_graph(adjacency, position)
    >>> image[1:4]
    'svg'
    """
    # check adjacency
    if adjacency is None:
        if position is None:
            raise ValueError("You must specify either adjacency or position.")
        else:
            n = position.shape[0]
            adjacency = sparse.csr_matrix((n, n)).astype(int)
    else:
        n = adjacency.shape[0]

    # node order
    if node_order is None:
        node_order = np.arange(n)

    # position
    if position is None:
        spring = Spring()
        position = spring.fit_transform(adjacency)

    # node colors
    node_colors = get_node_colors(n, labels, scores, membership, node_color, label_colors)

    # node sizes
    if node_weights is None:
        node_weights = adjacency.T.dot(np.ones(n))
    node_sizes = get_node_sizes(node_weights, node_size, node_size_min, node_size_max, display_node_weight)

    # node widths
    node_widths = get_node_widths(n, seeds, node_width, node_width_max)

    # rescaling
    position, width, height = rescale(position, width, height, margin, node_size, node_size_max, display_node_weight)

    if names is not None:
        text_length = np.max(np.array([len(str(name)) for name in names]))
        width += text_length * font_size * .5

    # scaling
    position *= scale
    height *= scale
    width *= scale

    svg = """<svg width="{}" height="{}" xmlns="http://www.w3.org/2000/svg">\n""".format(width, height)

    # edges
    if display_edges:
        adjacency_coo = sparse.coo_matrix(adjacency)

        if edge_color is None:
            if names is None:
                edge_color = 'black'
            else:
                edge_color = 'gray'

        if directed:
            svg += """<defs><marker id="arrow" markerWidth="10" markerHeight="10" refX="9" refY="3" orient="auto" >\n"""
            svg += """<path d="M0,0 L0,6 L9,3 z" fill="{}"/></marker></defs>\n""".format(edge_color)

        edge_colors, edge_order, edge_colors_residual = get_edge_colors(adjacency, edge_labels, edge_color,
                                                                        label_colors)
        edge_widths = get_edge_widths(adjacency_coo, edge_width, edge_width_min, edge_width_max, display_edge_weight)

        for ix in edge_order:
            i = adjacency_coo.row[ix]
            j = adjacency_coo.col[ix]
            color = edge_colors[ix]
            if directed:
                svg += svg_edge_directed(pos_1=position[i], pos_2=position[j], edge_width=edge_widths[ix],
                                         edge_color=color, node_size=node_sizes[j])
            else:
                svg += svg_edge(pos_1=position[i], pos_2=position[j],
                                edge_width=edge_widths[ix], edge_color=color)

        for i, j, color in edge_colors_residual:
            if directed:
                svg += svg_edge_directed(pos_1=position[i], pos_2=position[j], edge_width=edge_width,
                                         edge_color=color, node_size=node_sizes[j])
            else:
                svg += svg_edge(pos_1=position[i], pos_2=position[j],
                                edge_width=edge_width, edge_color=color)

    # nodes
    for i in node_order:
        if membership is None:
            svg += svg_node(position[i], node_sizes[i], node_colors[i], node_widths[i])
        else:
            if membership[i].nnz == 1:
                index = membership[i].indices[0]
                svg += svg_node(position[i], node_sizes[i], node_colors[index], node_widths[i])
            else:
                svg += svg_pie_chart_node(position[i], node_sizes[i], membership[i].todense(),
                                          node_colors, node_widths[i])

    # text
    if names is not None:
        for i in range(n):
            svg += svg_text(position[i] + node_sizes[i] + (margin_text, 0), names[i], font_size)
    svg += """</svg>\n"""

    if filename is not None:
        with open(filename + '.svg', 'w') as f:
            f.write(svg)

    return svg


def svg_digraph(adjacency: Optional[sparse.csr_matrix] = None, position: Optional[np.ndarray] = None,
                names: Optional[np.ndarray] = None,
                labels: Optional[Iterable] = None, scores: Optional[Iterable] = None,
                membership: Optional[sparse.csr_matrix] = None,
                seeds: Union[list, dict] = None, width: Optional[float] = 400, height: Optional[float] = 300,
                margin: float = 20, margin_text: float = 10, scale: float = 1, node_order: Optional[np.ndarray] = None,
                node_size: float = 7, node_size_min: float = 1, node_size_max: float = 20,
                display_node_weight: bool = False, node_weights: Optional[np.ndarray] = None, node_width: float = 1,
                node_width_max: float = 3, node_color: str = 'gray',
                display_edges: bool = True, edge_labels: Optional[list] = None,
                edge_width: float = 1, edge_width_min: float = 0.5,
                edge_width_max: float = 5, display_edge_weight: bool = True,
                edge_color: Optional[str] = None, label_colors: Optional[Iterable] = None,
                font_size: int = 12, filename: Optional[str] = None) -> str:
    """Return SVG image of a digraph.

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    position :
        Positions of the nodes.
    names :
        Names of the nodes.
    labels :
        Labels of the nodes (negative values mean no label).
    scores :
        Scores of the nodes (measure of importance).
    membership :
        Membership of the nodes (label distribution).
    seeds :
        Nodes to be highlighted (if dict, only keys are considered).
    width :
        Width of the image.
    height :
        Height of the image.
    margin :
        Margin of the image.
    margin_text :
        Margin between node and text.
    scale :
        Multiplicative factor on the dimensions of the image.
    node_order :
        Order in which nodes are displayed.
    node_size :
        Size of nodes.
    node_size_min :
        Minimum size of a node.
    node_size_max :
        Maximum size of a node.
    display_node_weight :
        If ``True``, display node in-weights through node size.
    node_weights :
        Node weights (used only if **display_node_weight** is ``True``).
    node_width :
        Width of node circle.
    node_width_max :
        Maximum width of node circle.
    node_color :
        Default color of nodes (svg color).
    display_edges :
        If ``True``, display edges.
    edge_labels :
        Labels of the edges, as a list of tuples (source, destination, label)
    edge_width :
        Width of edges.
    edge_width_min :
        Minimum width of edges.
    edge_width_max :
        Maximum width of edges.
    display_edge_weight :
        If ``True``, display edge weights through edge widths.
    edge_color :
        Default color of edges (svg color).
    label_colors :
        Colors of the labels (svg color).
    font_size :
        Font size.
    filename :
        Filename for saving image (optional).

    Returns
    -------
    image : str
        SVG image.

    Example
    -------
    >>> from sknetwork.data import painters
    >>> graph = painters(True)
    >>> adjacency = graph.adjacency
    >>> position = graph.position
    >>> from sknetwork.visualization import svg_digraph
    >>> image = svg_graph(adjacency, position)
    >>> image[1:4]
    'svg'
    """
    return svg_graph(adjacency=adjacency, position=position, names=names, labels=labels, scores=scores,
                     membership=membership, seeds=seeds, width=width, height=height, margin=margin,
                     margin_text=margin_text, scale=scale, node_order=node_order, node_size=node_size,
                     node_size_min=node_size_min, node_size_max=node_size_max, display_node_weight=display_node_weight,
                     node_weights=node_weights, node_width=node_width, node_width_max=node_width_max,
                     node_color=node_color, display_edges=display_edges, edge_labels=edge_labels,
                     edge_width=edge_width, edge_width_min=edge_width_min, edge_width_max=edge_width_max,
                     display_edge_weight=display_edge_weight, edge_color=edge_color, label_colors=label_colors,
                     font_size=font_size, directed=True, filename=filename)


def svg_bigraph(biadjacency: sparse.csr_matrix,
                names_row: Optional[np.ndarray] = None, names_col: Optional[np.ndarray] = None,
                labels_row: Optional[Union[dict, np.ndarray]] = None,
                labels_col: Optional[Union[dict, np.ndarray]] = None,
                scores_row: Optional[Union[dict, np.ndarray]] = None,
                scores_col: Optional[Union[dict, np.ndarray]] = None,
                membership_row: Optional[sparse.csr_matrix] = None,
                membership_col: Optional[sparse.csr_matrix] = None,
                seeds_row: Union[list, dict] = None, seeds_col: Union[list, dict] = None,
                position_row: Optional[np.ndarray] = None, position_col: Optional[np.ndarray] = None,
                reorder: bool = True, width: Optional[float] = 400,
                height: Optional[float] = 300, margin: float = 20, margin_text: float = 3, scale: float = 1,
                node_size: float = 7, node_size_min: float = 1, node_size_max: float = 20,
                display_node_weight: bool = False,
                node_weights_row: Optional[np.ndarray] = None, node_weights_col: Optional[np.ndarray] = None,
                node_width: float = 1, node_width_max: float = 3,
                color_row: str = 'gray', color_col: str = 'gray', label_colors: Optional[Iterable] = None,
                display_edges: bool = True, edge_labels: Optional[list] = None, edge_width: float = 1,
                edge_width_min: float = 0.5, edge_width_max: float = 10, edge_color: str = 'black',
                display_edge_weight: bool = True,
                font_size: int = 12, filename: Optional[str] = None) -> str:
    """Return SVG image of a bigraph.

    Parameters
    ----------
    biadjacency :
        Biadjacency matrix of the graph.
    names_row :
        Names of the rows.
    names_col :
        Names of the columns.
    labels_row :
        Labels of the rows (negative values mean no label).
    labels_col :
        Labels of the columns (negative values mean no label).
    scores_row :
        Scores of the rows (measure of importance).
    scores_col :
        Scores of the columns (measure of importance).
    membership_row :
        Membership of the rows (label distribution).
    membership_col :
        Membership of the columns (label distribution).
    seeds_row :
        Rows to be highlighted (if dict, only keys are considered).
    seeds_col :
        Columns to be highlighted (if dict, only keys are considered).
    position_row :
        Positions of the rows.
    position_col :
        Positions of the columns.
    reorder :
        Use clustering to order nodes.
    width :
        Width of the image.
    height :
        Height of the image.
    margin :
        Margin of the image.
    margin_text :
        Margin between node and text.
    scale :
        Multiplicative factor on the dimensions of the image.
    node_size :
        Size of nodes.
    node_size_min :
        Minimum size of nodes.
    node_size_max :
        Maximum size of nodes.
    display_node_weight :
        If ``True``, display node weights through node size.
    node_weights_row :
        Weights of rows (used only if **display_node_weight** is ``True``).
    node_weights_col :
        Weights of columns (used only if **display_node_weight** is ``True``).
    node_width :
        Width of node circle.
    node_width_max :
        Maximum width of node circle.
    color_row :
        Default color of rows (svg color).
    color_col :
        Default color of cols (svg color).
    label_colors :
        Colors of the labels (svg color).
    display_edges :
        If ``True``, display edges.
    edge_labels :
        Labels of the edges, as a list of tuples (source, destination, label)
    edge_width :
        Width of edges.
    edge_width_min :
        Minimum width of edges.
    edge_width_max :
        Maximum width of edges.
    display_edge_weight :
        If ``True``, display edge weights through edge widths.
    edge_color :
        Default color of edges (svg color).
    font_size :
        Font size.
    filename :
        Filename for saving image (optional).

    Returns
    -------
    image : str
        SVG image.

    Example
    -------
    >>> from sknetwork.data import movie_actor
    >>> biadjacency = movie_actor()
    >>> from sknetwork.visualization import svg_bigraph
    >>> image = svg_bigraph(biadjacency)
    >>> image[1:4]
    'svg'
    """
    n_row, n_col = biadjacency.shape

    # node positions
    if position_row is None or position_col is None:
        position_row = np.zeros((n_row, 2))
        position_col = np.ones((n_col, 2))
        if reorder:
            bilouvain = BiLouvain()
            bilouvain.fit(biadjacency)
            index_row = np.argsort(bilouvain.labels_row_)
            index_col = np.argsort(bilouvain.labels_col_)
        else:
            index_row = np.arange(n_row)
            index_col = np.arange(n_col)
        position_row[index_row, 1] = np.arange(n_row)
        position_col[index_col, 1] = np.arange(n_col) + .5 * (n_row - n_col)
    position = np.vstack((position_row, position_col))

    # node colors
    if scores_row is not None and scores_col is not None:
        if isinstance(scores_row, dict):
            scores_row = np.array(list(scores_row.values()))
        if isinstance(scores_col, dict):
            scores_col = np.array(list(scores_col.values()))
        scores = np.hstack((scores_row, scores_col))
        score_min = np.min(scores)
        score_max = np.max(scores)
    else:
        score_min = None
        score_max = None

    colors_row = get_node_colors(n_row, labels_row, scores_row, membership_row, color_row, label_colors,
                                 score_min, score_max)
    colors_col = get_node_colors(n_col, labels_col, scores_col, membership_col, color_col, label_colors,
                                 score_min, score_max)

    # node sizes
    if node_weights_row is None:
        node_weights_row = biadjacency.dot(np.ones(n_col))
    if node_weights_col is None:
        node_weights_col = biadjacency.T.dot(np.ones(n_row))
    node_sizes_row, node_sizes_col = get_node_sizes_bipartite(node_weights_row, node_weights_col,
                                                              node_size, node_size_min, node_size_max,
                                                              display_node_weight)

    # node widths
    node_widths_row = get_node_widths(n_row, seeds_row, node_width, node_width_max)
    node_widths_col = get_node_widths(n_col, seeds_col, node_width, node_width_max)

    # rescaling
    if not width and not height:
        raise ValueError("You must specify either the width or the height of the image.")
    position, width, height = rescale(position, width, height, margin, node_size, node_size_max, display_node_weight)

    # node names
    if names_row is not None:
        text_length = np.max(np.array([len(str(name)) for name in names_row]))
        position[:, 0] += text_length * font_size * .5
        width += text_length * font_size * .5
    if names_col is not None:
        text_length = np.max(np.array([len(str(name)) for name in names_col]))
        width += text_length * font_size * .5

    # scaling
    position *= scale
    height *= scale
    width *= scale
    position_row = position[:n_row]
    position_col = position[n_row:]

    svg = """<svg width="{}" height="{}"  xmlns="http://www.w3.org/2000/svg">\n""".format(width, height)

    # edges
    if display_edges:
        biadjacency_coo = sparse.coo_matrix(biadjacency)

        if edge_color is None:
            if names_row is None and names_col is None:
                edge_color = 'black'
            else:
                edge_color = 'gray'

        edge_colors, edge_order, edge_colors_residual = get_edge_colors(biadjacency, edge_labels, edge_color,
                                                                        label_colors)
        edge_widths = get_edge_widths(biadjacency_coo, edge_width, edge_width_min, edge_width_max, display_edge_weight)

        for ix in edge_order:
            i = biadjacency_coo.row[ix]
            j = biadjacency_coo.col[ix]
            color = edge_colors[ix]
            svg += svg_edge(pos_1=position_row[i], pos_2=position_col[j], edge_width=edge_widths[ix], edge_color=color)

        for i, j, color in edge_colors_residual:
            svg += svg_edge(pos_1=position_row[i], pos_2=position_col[j], edge_width=edge_width, edge_color=color)

    # nodes
    for i in range(n_row):
        if membership_row is None:
            svg += svg_node(position_row[i], node_sizes_row[i], colors_row[i], node_widths_row[i])
        else:
            if membership_row[i].nnz == 1:
                index = membership_row[i].indices[0]
                svg += svg_node(position_row[i], node_sizes_row[i], colors_row[index], node_widths_row[i])
            else:
                svg += svg_pie_chart_node(position_row[i], node_sizes_row[i], membership_row[i].todense(),
                                          colors_row, node_widths_row[i])

    for i in range(n_col):
        if membership_col is None:
            svg += svg_node(position_col[i], node_sizes_col[i], colors_col[i], node_widths_col[i])
        else:
            if membership_col[i].nnz == 1:
                index = membership_col[i].indices[0]
                svg += svg_node(position_col[i], node_sizes_col[i], colors_col[index], node_widths_col[i])
            else:
                svg += svg_pie_chart_node(position_col[i], node_sizes_col[i], membership_col[i].todense(),
                                          colors_col, node_widths_col[i])
    # text
    if names_row is not None:
        for i in range(n_row):
            svg += svg_text(position_row[i] - (margin_text + node_sizes_row[i], 0), names_row[i], font_size, True)
    if names_col is not None:
        for i in range(n_col):
            svg += svg_text(position_col[i] + (margin_text + node_sizes_col[i], 0), names_col[i], font_size)
    svg += """</svg>\n"""

    if filename is not None:
        with open(filename + '.svg', 'w') as f:
            f.write(svg)

    return svg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mar, 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from functools import partial
from multiprocessing import Pool
from typing import Union, Optional

import numpy as np
from scipy import sparse

from sknetwork.classification.base import BaseClassifier, BaseBiClassifier
from sknetwork.linalg.normalization import normalize
from sknetwork.ranking.base import BaseRanking
from sknetwork.utils.check import check_seeds, check_labels, check_n_jobs
from sknetwork.utils.format import bipartite2undirected
from sknetwork.utils.seeds import stack_seeds
from sknetwork.utils.verbose import VerboseMixin


class RankClassifier(BaseClassifier, VerboseMixin):
    """Generic class for ranking based classifiers.

    Parameters
    ----------
    algorithm :
        Which ranking algorithm to use.
    n_jobs :
        If positive, number of parallel jobs allowed (-1 means maximum number).
        If ``None``, no parallel computations are made.
    verbose :
        Verbose mode.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node.
    membership_ : sparse.csr_matrix
        Membership matrix (labels = columns).
    """
    def __init__(self, algorithm: BaseRanking, n_jobs: Optional[int] = None, verbose: bool = False):
        super(RankClassifier, self).__init__()
        VerboseMixin.__init__(self, verbose)

        self.algorithm = algorithm
        self.n_jobs = check_n_jobs(n_jobs)
        self.verbose = verbose

    @staticmethod
    def _process_seeds(labels_seeds: np.ndarray) -> list:
        """Make one-vs-all seed labels from seeds.

        Parameters
        ----------
        labels_seeds

        Returns
        -------
        List of seeds vectors.
        """
        seeds_all = []
        classes, _ = check_labels(labels_seeds)

        for label in classes:
            seeds = np.array(labels_seeds == label).astype(int)
            seeds_all.append(seeds)

        return seeds_all

    @staticmethod
    def _process_scores(scores: np.ndarray) -> np.ndarray:
        """Post-processing of the membership matrix.

        Parameters
        ----------
        scores
            Matrix of scores, shape number of nodes x number of labels.

        Returns
        -------
        scores: np.ndarray
        """
        return scores

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray], seeds: Union[np.ndarray, dict]) -> 'RankClassifier':
        """Fit algorithm to the data.

        Parameters
        ----------
        adjacency:
            Adjacency matrix of the graph.
        seeds:
            Seed nodes (labels as dictionary or array; negative values ignored).

        Returns
        -------
        self: :class:`RankClassifier`
        """
        n = adjacency.shape[0]
        seeds_labels = check_seeds(seeds, n).astype(int)
        classes, n_classes = check_labels(seeds_labels)

        seeds_all = self._process_seeds(seeds_labels)
        local_function = partial(self.algorithm.fit_transform, adjacency)
        with Pool(self.n_jobs) as pool:
            scores = np.array(pool.map(local_function, seeds_all))
        scores = scores.T

        scores = self._process_scores(scores)
        scores = normalize(scores)

        membership = sparse.coo_matrix(scores)
        membership.col = classes[membership.col]

        labels = np.argmax(scores, axis=1)
        self.labels_ = classes[labels]
        self.membership_ = sparse.csr_matrix(membership, shape=(n, np.max(seeds_labels) + 1))

        return self


class RankBiClassifier(RankClassifier, BaseBiClassifier):
    """Generic class for ranking based classifiers on bipartite graphs."""
    def __init__(self, algorithm: BaseRanking, n_jobs: Optional[int] = None, verbose: bool = False):
        super(RankBiClassifier, self).__init__(algorithm=algorithm, n_jobs=n_jobs, verbose=verbose)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray],
            seeds_row: Union[np.ndarray, dict], seeds_col: Union[np.ndarray, dict, None] = None) -> 'RankClassifier':
        """Compute labels.

        Parameters
        ----------
        biadjacency :
            Biadjacency matrix of the graph.
        seeds_row :
            Seed rows (labels as dictionary or array; negative values ignored).
        seeds_col :
            Seed columns (optional). Same format.

        Returns
        -------
        self: :class:`BiPageRankClassifier`
        """
        n_row, n_col = biadjacency.shape
        labels = stack_seeds(n_row, n_col, seeds_row, seeds_col)
        adjacency = bipartite2undirected(biadjacency)
        RankClassifier.fit(self, adjacency, labels)

        self.labels_row_ = self.labels_[:n_row]
        self.labels_col_ = self.labels_[n_row:]
        self.labels_ = self.labels_row_
        self.membership_row_ = self.membership_[:n_row]
        self.membership_col_ = self.membership_[n_row:]
        self.membership_ = self.membership_row_

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""""tests for metrics.py"""
import unittest

from sknetwork.data import house
from sknetwork.path import diameter


class TestMetrics(unittest.TestCase):

    def test_diameter(self):
        adjacency = house()
        with self.assertRaises(ValueError):
            diameter(adjacency, 2.5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov, 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from abc import ABC

import numpy as np
from scipy import sparse

from sknetwork.linalg.normalization import normalize
from sknetwork.utils.base import Algorithm
from sknetwork.utils.membership import membership_matrix


class BaseClustering(Algorithm, ABC):
    """Base class for clustering algorithms.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node.
    membership_ : sparse.csr_matrix
        Membership matrix.
    adjacency_ : sparse.csr_matrix
        Adjacency matrix between clusters.
    """
    def __init__(self, sort_clusters: bool = True, return_membership: bool = False, return_aggregate: bool = False):
        self.sort_clusters = sort_clusters
        self.return_membership = return_membership
        self.return_aggregate = return_aggregate

        self.labels_ = None
        self.membership_ = None
        self.adjacency_ = None

    def fit_transform(self, *args, **kwargs) -> np.ndarray:
        """Fit algorithm to the data and return the labels. Same parameters as the ``fit`` method.

        Returns
        -------
        labels : np.ndarray
            Labels.
        """
        self.fit(*args, **kwargs)
        return self.labels_

    def _secondary_outputs(self, adjacency):
        """Compute different variables from labels_."""
        if self.return_membership or self.return_aggregate:
            if np.issubdtype(adjacency.data.dtype, np.bool_):
                adjacency = adjacency.astype(float)
            membership = membership_matrix(self.labels_)
            if self.return_membership:
                self.membership_ = normalize(adjacency.dot(membership))
            if self.return_aggregate:
                self.adjacency_ = sparse.csr_matrix(membership.T.dot(adjacency.dot(membership)))
        return self


class BaseBiClustering(BaseClustering, ABC):
    """Base class for clustering algorithms.

    Attributes
    ----------
    labels_ : np.ndarray
        Labels of the rows.
    labels_row_ : np.ndarray
        Labels of the rows (copy of **labels_**).
    labels_col_ : np.ndarray
        Labels of the columns.
    membership_ : sparse.csr_matrix
        Membership matrix of the rows (copy of **membership_**).
    membership_row_ : sparse.csr_matrix
        Membership matrix of the rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of the columns. Only valid if **cluster_both** = `True`.
    biadjacency_ : sparse.csr_matrix
        Biadjacency matrix of the aggregate graph between clusters.
    """

    def __init__(self, sort_clusters: bool = True, return_membership: bool = False, return_aggregate: bool = False):
        super(BaseBiClustering, self).__init__(sort_clusters=sort_clusters, return_membership=return_membership,
                                               return_aggregate=return_aggregate)

        self.labels_row_ = None
        self.labels_col_ = None
        self.membership_row_ = None
        self.membership_col_ = None
        self.biadjacency_ = None

    def _split_vars(self, n_row):
        """Split labels_ into labels_row_ and labels_col_"""
        self.labels_row_ = self.labels_[:n_row]
        self.labels_col_ = self.labels_[n_row:]
        self.labels_ = self.labels_row_
        return self

    def _secondary_outputs(self, biadjacency):
        """Compute different variables from labels_."""
        if self.return_membership or self.return_aggregate:
            if np.issubdtype(biadjacency.data.dtype, np.bool_):
                biadjacency = biadjacency.astype(float)

        if self.return_membership:
            membership_row = membership_matrix(self.labels_row_)
            membership_col = membership_matrix(self.labels_col_)
            self.membership_row_ = normalize(biadjacency.dot(membership_col))
            self.membership_col_ = normalize(biadjacency.T.dot(membership_row))
            self.membership_ = self.membership_row_

        if self.return_aggregate:
            membership_row = membership_matrix(self.labels_row_)
            membership_col = membership_matrix(self.labels_col_)
            biadjacency_ = sparse.csr_matrix(membership_row.T.dot(biadjacency))
            biadjacency_ = biadjacency_.dot(membership_col)
            self.biadjacency_ = biadjacency_
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May, 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Union, Optional

import numpy as np
from scipy import sparse

from sknetwork.path.shortest_path import distance


def diameter(adjacency: Union[sparse.csr_matrix, np.ndarray], n_sources: Optional[Union[int, float]] = None,
             n_jobs: Optional[int] = None) -> int:
    """Lower bound on the diameter of a graph which is the length of the longest shortest path between two nodes.

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    n_sources :
        Number of node sources to use for approximation.

        * If None, compute exact diameter.
        * If int, sample n_sample source nodes at random.
        * If float, sample (n_samples * n) source nodes at random.
    n_jobs :
        If an integer value is given, denotes the number of workers to use (-1 means the maximum number will be used).
        If ``None``, no parallel computations are made.

    Returns
    -------
    diameter : int

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> d_exact = diameter(adjacency)
    >>> d_exact
    2
    >>> d_approx = diameter(adjacency, 2)
    >>> d_approx <= d_exact
    True
    >>> d_approx = diameter(adjacency, 0.5)
    >>> d_approx <= d_exact
    True

    Notes
    -----
    This is a basic implementation that computes distances between nodes and returns the maximum.
    """
    n = adjacency.shape[0]
    if n_sources is None or n_sources == n:
        sources = np.arange(n)
    else:
        if np.issubdtype(type(n_sources), np.floating) and n_sources < 1.:
            n_sources = int(n_sources * n)
        if np.issubdtype(type(n_sources), np.integer) and n_sources <= n:
            sources = np.random.choice(n, n_sources, replace=False)
        else:
            raise ValueError("n_sources must be either None, an integer smaller than the number of nodes or a float"
                             "smaller than 1.")
    dists = distance(adjacency, sources, method='D', return_predecessors=False, n_jobs=n_jobs).astype(int)
    return dists.max()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from abc import ABC

import numpy as np
from scipy import sparse
from scipy.spatial import cKDTree

from sknetwork.utils.base import Algorithm
from sknetwork.utils.format import directed2undirected
from sknetwork.utils.knn1d import knn1d


class BaseTransformer(Algorithm, ABC):
    """Base class for transformers."""
    def __init__(self, undirected: bool = False):
        self.undirected = undirected

        self.adjacency_ = None

    def fit_transform(self, x: np.ndarray) -> sparse.csr_matrix:
        """Fit algorithm to the data and return the computed adjacency.

        Parameters
        ----------
        x: np.ndarray
            Input data.

        Returns
        -------
        adjacency : sparse.csr_matrix
        """
        self.fit(x)
        return self.adjacency_

    def make_undirected(self):
        """Modifies the adjacency to match desired constrains."""
        if self.adjacency_ is not None and self.undirected:
            dtype = self.adjacency_.dtype
            self.adjacency_ = directed2undirected(self.adjacency_, weighted=False).astype(dtype)

        return self


class KNNDense(BaseTransformer):
    """Extract adjacency from vector data through k-nearest-neighbor search with KD-Tree.

    Parameters
    ----------
    n_neighbors :
        Number of neighbors for each sample in the transformed sparse graph.
    undirected :
        As the nearest neighbor relationship is not symmetric, the graph is directed by default.
        Setting this parameter to ``True`` forces the algorithm to return undirected graphs.
    leaf_size :
        Leaf size passed to KDTree.
        This can affect the speed of the construction and query, as well as the memory required to store the tree.
    p :
        Which Minkowski p-norm to use. 1 is the sum-of-absolute-values “Manhattan” distance,
        2 is the usual Euclidean distance infinity is the maximum-coordinate-difference distance.
        A finite large p may cause a ValueError if overflow can occur.
    eps :
        Return approximate nearest neighbors; the k-th returned value is guaranteed to be no further than (1+tol_nn)
        times the distance to the real k-th nearest neighbor.
    n_jobs :
        Number of jobs to schedule for parallel processing. If -1 is given all processors are used.

    Attributes
    ----------
    adjacency_ :
        Adjacency matrix of the graph.

    References
    ----------
    Maneewongvatana, S., & Mount, D. M. (1999, December). It’s okay to be skinny, if your friends are fat.
    In Center for Geometric Computing 4th Annual Workshop on Computational Geometry (Vol. 2, pp. 1-8).
    """
    def __init__(self, n_neighbors: int = 5, undirected: bool = False, leaf_size: int = 16, p=2, eps: float = 0.01,
                 n_jobs=1):
        super(KNNDense, self).__init__(undirected)

        self.n_neighbors = n_neighbors
        self.leaf_size = leaf_size
        self.p = p
        self.eps = eps
        self.n_jobs = n_jobs

    def fit(self, x: np.ndarray) -> 'KNNDense':
        """Fit algorithm to the data.

        Parameters
        ----------
        x :
            Data to transform into adjacency.

        Returns
        -------
        self : :class:`KNNDense`
        """
        tree = cKDTree(x, self.leaf_size)
        _, neighbors = tree.query(x, self.n_neighbors + 1, self.eps, self.p, n_jobs=self.n_jobs)

        n: int = x.shape[0]
        indptr: np.ndarray = np.arange(n + 1) * (self.n_neighbors + 1)
        indices: np.ndarray = neighbors.reshape(-1)
        data = np.ones(indices.shape[0], dtype=bool)

        self.adjacency_ = sparse.csr_matrix((data, indices, indptr))
        self.make_undirected()
        self.adjacency_.setdiag(0)

        return self


class CNNDense(BaseTransformer):
    """Extract adjacency from vector data through component-wise k-nearest-neighbor search.
    KNN is applied independently on each column of the input matrix.

    Parameters
    ----------
    n_neighbors :
        Number of neighbors per dimension.
    undirected :
        As the nearest neighbor relationship is not symmetric, the graph is directed by default.
        Setting this parameter to ``True`` forces the algorithm to return undirected graphs.

    Attributes
    ----------
    adjacency_ :
        Adjacency matrix of the  graph.
    """
    def __init__(self, n_neighbors: int = 1, undirected: bool = False):
        super(CNNDense, self).__init__(undirected)

        self.n_neighbors = n_neighbors

    def fit(self, x: np.ndarray) -> 'CNNDense':
        """Fit algorithm to the data.

        Parameters
        ----------
        x:
            Data to transform into adjacency.

        Returns
        -------
        self: :class:`CNNDense`
        """
        rows, cols = [], []
        for j in range(x.shape[1]):
            row, col = knn1d(x[:, j].astype(np.float32), self.n_neighbors)
            rows += row
            cols += col

        rows = np.array(rows)
        cols = np.array(cols)
        data = np.ones(cols.shape[0], dtype=bool)

        self.adjacency_ = sparse.csr_matrix((data, (rows, cols)))
        self.make_undirected()

        return self



#!/usr/bin/env python3
# coding: utf-8
"""
Created on May, 2020
@author: Thomas Bonald <tbonald@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse

from sknetwork.classification.propagation import Propagation
from sknetwork.clustering.base import BaseClustering, BaseBiClustering
from sknetwork.utils.check import check_format
from sknetwork.utils.format import bipartite2undirected


class PropagationClustering(BaseClustering, Propagation):
    """Clustering by label propagation.

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_iter : int
        Maximum number of iterations (-1 for infinity).
    node_order : str
        * `'random'`: node labels are updated in random order.
        * `'increasing'`: node labels are updated by increasing order of weight.
        * `'decreasing'`: node labels are updated by decreasing order of weight.
        * Otherwise, node labels are updated by index order.
    weighted : bool
        If ``True``, the vote of each neighbor is proportional to the edge weight.
        Otherwise, all votes have weight 1.
    sort_clusters :
            If ``True``, sort labels in decreasing order of cluster size.
    return_membership :
            If ``True``, return the membership matrix of nodes to each cluster (soft clustering).
    return_aggregate :
            If ``True``, return the adjacency matrix of the graph between clusters.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node.
    membership_ : sparse.csr_matrix
        Membership matrix (columns = labels).

    Example
    -------
    >>> from sknetwork.clustering import PropagationClustering
    >>> from sknetwork.data import karate_club
    >>> propagation = PropagationClustering()
    >>> graph = karate_club(metadata=True)
    >>> adjacency = graph.adjacency
    >>> labels = propagation.fit_transform(adjacency)
    >>> len(set(labels))
    2

    References
    ----------
    Raghavan, U. N., Albert, R., & Kumara, S. (2007).
    `Near linear time algorithm to detect community structures in large-scale networks.
    <https://arxiv.org/pdf/0709.2938.pdf>`_
    Physical review E, 76(3), 036106.
    """
    def __init__(self, n_iter: int = 5, node_order: str = 'decreasing', weighted: bool = True,
                 sort_clusters: bool = True, return_membership: bool = True, return_aggregate: bool = True):
        Propagation.__init__(self, n_iter, node_order, weighted)
        BaseClustering.__init__(self, sort_clusters, return_membership, return_aggregate)

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'PropagationClustering':
        """Clustering by label propagation.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`PropagationClustering`
        """
        Propagation.fit(self, adjacency)
        _, labels = np.unique(self.labels_, return_inverse=True)

        self.labels_ = labels
        self._secondary_outputs(adjacency)

        return self


class BiPropagationClustering(PropagationClustering, BaseBiClustering):
    """Clustering by label propagation in bipartite graphs.

    * Bigraphs

    Parameters
    ----------
    n_iter :
        Maximum number of iteration (-1 for infinity).
    sort_clusters :
            If ``True``, sort labels in decreasing order of cluster size.
    return_membership :
            If ``True``, return the membership matrix of nodes to each cluster (soft clustering).
    return_aggregate :
        If ``True``, return the biadjacency matrix of the graph between clusters.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each row.
    labels_row_ : np.ndarray
        Label of each row (copy of **labels_**).
    labels_col_ : np.ndarray
        Label of each column.
    membership_ : sparse.csr_matrix
        Membership matrix of rows.
    membership_row_ : sparse.csr_matrix
        Membership matrix of rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of columns.

    Example
    -------
    >>> from sknetwork.clustering import BiPropagationClustering
    >>> from sknetwork.data import movie_actor
    >>> bipropagation = BiPropagationClustering()
    >>> graph = movie_actor(metadata=True)
    >>> biadjacency = graph.biadjacency
    >>> len(bipropagation.fit_transform(biadjacency))
    15
    >>> len(bipropagation.labels_col_)
    16
    """
    def __init__(self, n_iter: int = 5, node_order: str = 'decreasing', weighted: bool = True,
                 sort_clusters: bool = True, return_membership: bool = True, return_aggregate: bool = True):
        super(BiPropagationClustering, self).__init__(n_iter=n_iter, node_order=node_order, weighted=weighted,
                                                      sort_clusters=sort_clusters, return_membership=return_membership,
                                                      return_aggregate=return_aggregate)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'BiPropagationClustering':
        """Clustering.

        Parameters
        ----------
        biadjacency :
            Biadjacency matrix of the graph.

        Returns
        -------
        self: :class:`BiPropagationClustering`
        """
        n_row, n_col = biadjacency.shape
        biadjacency = check_format(biadjacency)
        adjacency = bipartite2undirected(biadjacency)

        propagation = PropagationClustering(self.n_iter, self.node_order, self.weighted)

        self.labels_ = propagation.fit_transform(adjacency)
        self._split_vars(n_row)
        self._secondary_outputs(biadjacency)

        return self



"""embedding module"""
from sknetwork.embedding.base import BaseEmbedding, BaseBiEmbedding
from sknetwork.embedding.force_atlas import ForceAtlas
from sknetwork.embedding.louvain_embedding import BiLouvainEmbedding, LouvainEmbedding
from sknetwork.embedding.metrics import cosine_modularity
from sknetwork.embedding.spectral import Spectral, BiSpectral, LaplacianEmbedding
from sknetwork.embedding.spring import Spring
from sknetwork.embedding.svd import SVD, GSVD, PCA



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Jun 28, 2019
@author: Quentin Lutz <qlutz@enst.fr>
"""
import inspect


class Algorithm:
    """Base class for all algorithms."""
    def __repr__(self):
        # parameters not to display
        arg_black_list = ['self', 'random_state', 'verbose']
        output = self.__class__.__name__ + '('
        signature = inspect.signature(self.__class__.__init__)
        arguments = [arg.name for arg in signature.parameters.values() if arg.name not in arg_black_list]
        for p in arguments:
            try:
                val = self.__dict__[p]
            except KeyError:
                continue
            if type(val) == str:
                val = "'" + val + "'"
            else:
                val = str(val)
            output += p + '=' + val + ', '
        if output[-1] != '(':
            return output[:-2] + ')'
        else:
            return output + ')'

    def fit(self, *args, **kwargs):
        """Fit Algorithm to the data."""
        raise NotImplementedError



#!/usr/bin/env python3
# coding: utf-8
"""
Created on July 10 2019
@author: Nathan De Lara <ndelara@enst.fr>
"""


def auto_solver(nnz: int, threshold: int = 1e4) -> str:
    """Recommend a solver for SVD or Eigenvalue decomposition depending of the size of the input matrix.
    Halko's randomized method is returned for big matrices and Lanczos for small ones.

    Parameters
    ----------
    nnz: int
        Number of non-zero entries of the matrix to decompose.
    threshold: int
        Threshold beyond which randomized methods are applied.

    Examples
    --------
    >>> auto_solver(1000)
    'lanczos'
    >>> auto_solver(100000)
    'halko'

    Returns
    -------
    solver name: str
        'halko' or ' lanczos'
    """
    if nnz > threshold:
        return 'halko'
    else:
        return 'lanczos'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for visualization/dendrograms.py"""

import unittest

import numpy as np

from sknetwork.data.toy_graphs import karate_club, painters
from sknetwork.hierarchy import Paris
from sknetwork.visualization.dendrograms import svg_dendrogram, svg_dendrogram_top


# noinspection DuplicatedCode
class TestVisualization(unittest.TestCase):

    def test_undirected(self):
        adjacency = karate_club()
        paris = Paris()
        dendrogram = paris.fit_transform(adjacency)
        image = svg_dendrogram(dendrogram)
        self.assertEqual(image[1:4], 'svg')
        n = adjacency.shape[0]
        image = svg_dendrogram(dendrogram, names=np.arange(n), width=200, height=200, margin=10, margin_text=5, scale=3,
                               n_clusters=3, color='green', colors=['red', 'blue'], font_size=14, reorder=True,
                               rotate=True)
        self.assertEqual(image[1:4], 'svg')
        image = svg_dendrogram(dendrogram, names=np.arange(n), width=200, height=200, margin=10, margin_text=5, scale=3,
                               n_clusters=3, color='green', colors={0: 'red', 1: 'blue'}, font_size=14, reorder=False,
                               rotate=True)
        self.assertEqual(image[1:4], 'svg')
        svg_dendrogram_top(dendrogram, names=np.arange(n), width=200, height=200, margin=10, margin_text=5, scale=3,
                           n_clusters=3, color='green', colors=np.array(['red', 'black', 'blue']), font_size=14,
                           reorder=False, rotate_names=True, line_width=0.1)

    def test_directed(self):
        graph = painters(True)
        adjacency = graph.adjacency
        names = graph.names
        paris = Paris()
        dendrogram = paris.fit_transform(adjacency)
        image = svg_dendrogram(dendrogram)
        self.assertEqual(image[1:4], 'svg')
        image = svg_dendrogram(dendrogram, names=names, width=200, height=200, margin=10, margin_text=5, scale=3,
                               n_clusters=3, color='green', font_size=14, reorder=True, rotate=True)
        self.assertEqual(image[1:4], 'svg')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for load.py"""

import tempfile
import unittest
import warnings

import numpy as np

# has to specify the exact file to avoid nosetests error on full tests
from sknetwork.data.load import load_netset, load_konect, clear_data_home, save, load
from sknetwork.data.toy_graphs import house, star_wars
from sknetwork.utils.timeout import TimeOut


class TestLoader(unittest.TestCase):

    def test_netset(self):
        tmp_data_dir = tempfile.gettempdir() + '/stub'
        clear_data_home(tmp_data_dir)
        try:
            graph = load_netset('stub', tmp_data_dir)
        except:  # pragma: no cover
            warnings.warn('Could not reach Telecom Graphs. Corresponding test has not been performed.', RuntimeWarning)
            return
        n = 2
        self.assertEqual(graph.adjacency.shape, (n, n))
        self.assertEqual(len(graph.names), n)
        clear_data_home(tmp_data_dir)

    def test_invalid_netset(self):
        tmp_data_dir = tempfile.gettempdir() + '/stub'
        clear_data_home(tmp_data_dir)
        try:
            with self.assertRaises(ValueError):
                load_netset('junk', tmp_data_dir)

        except:  # pragma: no cover
            warnings.warn('Could not reach Telecom Graphs. Corresponding test has not been performed.', RuntimeWarning)
            return
        load_netset()

    def test_konect(self):
        tmp_data_dir = tempfile.gettempdir() + '/moreno_crime'
        clear_data_home(tmp_data_dir)
        try:
            with TimeOut(2):
                data = load_konect('moreno_crime', tmp_data_dir)
        except (TimeoutError, RuntimeError):  # pragma: no cover
            warnings.warn('Could not reach Konect. Corresponding test has not been performed.', RuntimeWarning)
            return
        self.assertEqual(data.biadjacency.shape[0], 829)
        self.assertEqual(data.name.shape[0], 829)
        data = load_konect('moreno_crime', tmp_data_dir)
        self.assertEqual(data.biadjacency.shape[0], 829)
        self.assertEqual(data.name.shape[0], 829)
        clear_data_home(tmp_data_dir)

        tmp_data_dir = tempfile.gettempdir() + '/ego-facebook'
        try:
            with TimeOut(2):
                data = load_konect('ego-facebook', tmp_data_dir)
        except (TimeoutError, RuntimeError):  # pragma: no cover
            warnings.warn('Could not reach Konect. Corresponding test has not been performed.', RuntimeWarning)
            return
        self.assertEqual(data.adjacency.shape[0], 2888)
        data = load_konect('ego-facebook', tmp_data_dir)
        self.assertEqual(data.adjacency.shape[0], 2888)
        clear_data_home(tmp_data_dir)

    def test_invalid_konect(self):
        tmp_data_dir = tempfile.gettempdir() + '/stub'
        clear_data_home(tmp_data_dir)
        try:
            with TimeOut(4):
                with self.assertRaises(ValueError):
                    load_konect('junk', tmp_data_dir)
                with self.assertRaises(ValueError):
                    load_konect('', tmp_data_dir)
        except (TimeoutError, RuntimeError):  # pragma: no cover
            warnings.warn('Could not reach Konect. Corresponding test has not been performed.', RuntimeWarning)
            return

    def test_save_load(self):
        data = house()
        tmp_data_dir = tempfile.gettempdir() + '/stub'
        save(tmp_data_dir + '/house', data)
        loaded_data = load(tmp_data_dir + '/house')
        self.assertTrue(np.allclose(data.data, loaded_data.adjacency.data))

        data = star_wars()
        save(tmp_data_dir + '/star_wars', data)
        loaded_data = load(tmp_data_dir + '/star_wars')
        self.assertTrue(np.allclose(data.data, loaded_data.biadjacency.data))

        data = star_wars(metadata=True)
        save(tmp_data_dir + '/star_wars', data)
        loaded_data = load(tmp_data_dir + '/star_wars')
        self.assertTrue(np.allclose(data.biadjacency.data, loaded_data.biadjacency.data))
        self.assertEqual(data.names_col[0], loaded_data.names_col[0])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on March 2019
@author: Thomas Bonald <bonald@enst.fr>
"""
import numpy as np
from scipy import sparse
from sknetwork.hierarchy.paris import AggregateGraph

from sknetwork.utils.check import check_format, check_probs, check_square
from sknetwork.utils.check import check_min_size, check_min_nnz
from sknetwork.utils.format import directed2undirected


def _instanciate_vars(adjacency: sparse.csr_matrix, weights: str = 'uniform'):
    """Initialize standard variables for metrics."""
    n = adjacency.shape[0]
    weights_row = check_probs(weights, adjacency)
    weights_col = check_probs(weights, adjacency.T)
    sym_adjacency = directed2undirected(adjacency)

    aggregate_graph = AggregateGraph(weights_row, weights_col, sym_adjacency.data.astype(np.float),
                                     sym_adjacency.indices, sym_adjacency.indptr)

    height = np.zeros(n - 1)
    cluster_weight = np.zeros(n - 1)
    edge_sampling = np.zeros(n - 1)

    return aggregate_graph, height, cluster_weight, edge_sampling, weights_row, weights_col


def dasgupta_cost(adjacency: sparse.csr_matrix, dendrogram: np.ndarray, weights: str = 'uniform',
                  normalized: bool = False) -> float:
    """Dasgupta's cost of a hierarchy.

    * Graphs
    * Digraphs

    Expected size (weights = ``'uniform'``) or expected weight (weights = ``'degree'``) of the cluster induced by
    random edge sampling (closest ancestor of the two nodes in the hierarchy).

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    dendrogram :
        Dendrogram.
    weights :
        Weights of nodes.
        ``'degree'`` or ``'uniform'`` (default).
    normalized :
        If ``True``, normalized cost (between 0 and 1).

    Returns
    -------
    cost : float
        Cost.

    Example
    -------
    >>> from sknetwork.hierarchy import dasgupta_score, Paris
    >>> from sknetwork.data import house
    >>> paris = Paris()
    >>> adjacency = house()
    >>> dendrogram = paris.fit_transform(adjacency)
    >>> cost = dasgupta_cost(adjacency, dendrogram)
    >>> np.round(cost, 2)
    3.33

    References
    ----------
    Dasgupta, S. (2016). A cost function for similarity-based hierarchical clustering.
    Proceedings of ACM symposium on Theory of Computing.
    """
    adjacency = check_format(adjacency)
    check_square(adjacency)

    n = adjacency.shape[0]
    check_min_size(n, 2)

    aggregate_graph, height, edge_sampling, cluster_weight, _, _ = _instanciate_vars(adjacency, weights)

    for t in range(n - 1):
        i = int(dendrogram[t][0])
        j = int(dendrogram[t][1])
        if i >= n and height[i - n] == dendrogram[t][2]:
            edge_sampling[t] = edge_sampling[i - n]
            edge_sampling[i - n] = 0
        elif j >= n and height[j - n] == dendrogram[t][2]:
            edge_sampling[t] = edge_sampling[j - n]
            edge_sampling[j - n] = 0
        height[t] = dendrogram[t][2]
        if j in aggregate_graph.neighbors[i]:
            edge_sampling[t] += aggregate_graph.neighbors[i][j]
        cluster_weight[t] = aggregate_graph.cluster_out_weights[i] + aggregate_graph.cluster_out_weights[j] \
            + aggregate_graph.cluster_in_weights[i] + aggregate_graph.cluster_in_weights[j]
        aggregate_graph.merge(i, j)

    cost: float = edge_sampling.dot(cluster_weight) / 2

    if not normalized:
        if weights == 'degree':
            cost *= adjacency.data.sum()
        else:
            cost *= n

    return cost


def dasgupta_score(adjacency: sparse.csr_matrix, dendrogram: np.ndarray, weights: str = 'uniform') -> float:
    """Dasgupta's score of a hierarchy (quality metric, between 0 and 1).

    * Graphs
    * Digraphs

    Defined as 1 - normalized Dasgupta's cost.

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    dendrogram :
        Dendrogram.
    weights :
        Weights of nodes.
        ``'degree'`` or ``'uniform'`` (default).

    Returns
    -------
    score : float
        Score.

    Example
    -------
    >>> from sknetwork.hierarchy import dasgupta_score, Paris
    >>> from sknetwork.data import house
    >>> paris = Paris()
    >>> adjacency = house()
    >>> dendrogram = paris.fit_transform(adjacency)
    >>> score = dasgupta_score(adjacency, dendrogram)
    >>> np.round(score, 2)
    0.33

    References
    ----------
    Dasgupta, S. (2016). A cost function for similarity-based hierarchical clustering.
    Proceedings of ACM symposium on Theory of Computing.
    """
    return 1 - dasgupta_cost(adjacency, dendrogram, weights, normalized=True)


def tree_sampling_divergence(adjacency: sparse.csr_matrix, dendrogram: np.ndarray, weights: str = 'degree',
                             normalized: bool = True) -> float:
    """Tree sampling divergence of a hierarchy (quality metric).

    * Graphs
    * Digraphs

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    dendrogram :
        Dendrogram.
    weights :
        Weights of nodes.
        ``'degree'`` (default) or ``'uniform'``.
    normalized :
        If ``True``, normalized score (between 0 and 1).

    Returns
    -------
    score : float
        Score.

    Example
    -------
    >>> from sknetwork.hierarchy import tree_sampling_divergence, Paris
    >>> from sknetwork.data import house
    >>> paris = Paris()
    >>> adjacency = house()
    >>> dendrogram = paris.fit_transform(adjacency)
    >>> score = tree_sampling_divergence(adjacency, dendrogram)
    >>> np.round(score, 2)
    0.52

    References
    ----------
    Charpentier, B. & Bonald, T. (2019).
    `Tree Sampling Divergence: An Information-Theoretic Metric for
    Hierarchical Graph Clustering.
    <https://hal.telecom-paristech.fr/hal-02144394/document>`_
    Proceedings of IJCAI.
    """
    adjacency = check_format(adjacency)
    check_square(adjacency)
    check_min_nnz(adjacency.nnz, 1)
    adjacency = adjacency.astype(float)
    n = adjacency.shape[0]
    check_min_size(n, 2)

    adjacency.data /= adjacency.data.sum()

    aggregate_graph, height, cluster_weight, edge_sampling, weights_row, weights_col = _instanciate_vars(adjacency,
                                                                                                         weights)
    node_sampling = np.zeros(n - 1)

    for t in range(n - 1):
        i = int(dendrogram[t][0])
        j = int(dendrogram[t][1])
        if i >= n and height[i - n] == dendrogram[t][2]:
            edge_sampling[t] = edge_sampling[i - n]
            edge_sampling[i - n] = 0
            node_sampling[t] = node_sampling[i - n]
        elif j >= n and height[j - n] == dendrogram[t][2]:
            edge_sampling[t] = edge_sampling[j - n]
            edge_sampling[j - n] = 0
            node_sampling[t] = node_sampling[j - n]
        if j in aggregate_graph.neighbors[i]:
            edge_sampling[t] += aggregate_graph.neighbors[i][j]
        node_sampling[t] += aggregate_graph.cluster_out_weights[i] * aggregate_graph.cluster_in_weights[j] + \
            aggregate_graph.cluster_out_weights[j] * aggregate_graph.cluster_in_weights[i]
        height[t] = dendrogram[t][2]
        aggregate_graph.merge(i, j)

    index = np.where(edge_sampling)[0]
    score = edge_sampling[index].dot(np.log(edge_sampling[index] / node_sampling[index]))
    if normalized:
        inv_out_weights = sparse.diags(weights_row, shape=(n, n), format='csr')
        inv_out_weights.data = 1 / inv_out_weights.data
        inv_in_weights = sparse.diags(weights_col, shape=(n, n), format='csr')
        inv_in_weights.data = 1 / inv_in_weights.data
        sampling_ratio = inv_out_weights.dot(adjacency.dot(inv_in_weights))
        inv_out_weights.data = np.ones(len(inv_out_weights.data))
        inv_in_weights.data = np.ones(len(inv_in_weights.data))
        edge_sampling = inv_out_weights.dot(adjacency.dot(inv_in_weights))
        mutual_information = edge_sampling.data.dot(np.log(sampling_ratio.data))
        score /= mutual_information
    return score



"""ranking module"""
from sknetwork.ranking.base import BaseRanking
from sknetwork.ranking.betweenness import Betweenness
from sknetwork.ranking.closeness import Closeness
from sknetwork.ranking.diffusion import Diffusion, BiDiffusion, Dirichlet, BiDirichlet
from sknetwork.ranking.harmonic import Harmonic
from sknetwork.ranking.hits import HITS
from sknetwork.ranking.katz import Katz, BiKatz
from sknetwork.ranking.pagerank import PageRank, BiPageRank
from sknetwork.ranking.postprocess import top_k



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on July, 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
import numpy as np


def accuracy_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Accuracy: number of correctly labeled samples over total number of elements.
     In the case of binary classification, this is

    :math:`P = \\dfrac{TP + TN}{TP + TN + FP + FN}`.

    Parameters
    ----------
    y_true : np.ndarray
        True labels.
    y_pred : np.ndarray
        Predicted labels

    Returns
    -------
    precision : float
        A score between 0 and 1.

    Examples
    --------
    >>> import numpy as np
    >>> y_true = np.array([0, 0, 1, 1])
    >>> y_pred = np.array([0, 0, 0, 1])
    >>> accuracy_score(y_true, y_pred)
    0.75
    """
    return (y_true == y_pred).mean()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on November 12, 2019
@author: Quentin Lutz <qlutz@enst.fr>
"""
from functools import partial
from multiprocessing import Pool
from typing import Optional, Union, Iterable

import numpy as np
from scipy import sparse

from sknetwork.utils.check import check_n_jobs, is_symmetric


def distance(adjacency: sparse.csr_matrix, sources: Optional[Union[int, Iterable]] = None, method: str = 'D',
             return_predecessors: bool = False, unweighted: bool = False, n_jobs: Optional[int] = None):
    """Compute distances between nodes.

    * Graphs
    * Digraphs

    Based on SciPy (scipy.sparse.csgraph.shortest_path)

    Parameters
    ----------
    adjacency :
        The adjacency matrix of the graph
    sources :
        If specified, only compute the paths for the points at the given indices. Will not work with ``method =='FW'``.
    method :
        The method to be used.

        * ``'D'`` (Dijkstra),
        * ``'BF'`` (Bellman-Ford),
        * ``'J'`` (Johnson).
    return_predecessors :
        If ``True``, the size predecessor matrix is returned
    unweighted :
        If ``True``, the weights of the edges are ignored
    n_jobs :
        If an integer value is given, denotes the number of workers to use (-1 means the maximum number will be used).
        If ``None``, no parallel computations are made.

    Returns
    -------
    dist_matrix : np.ndarray
        The matrix of distances between graph nodes. ``dist_matrix[i,j]`` gives the shortest
        distance from point ``i`` to point ``j`` along the graph.
        If no path exists between nodes ``i`` and ``j``, then ``dist_matrix[i, j] = np.inf``.
    predecessors : np.ndarray, optional
        Returned only if ``return_predecessors == True``. The matrix of predecessors, which can be used to reconstruct
        the shortest paths. Row i of the predecessor matrix contains information on the shortest paths from point ``i``:
        each entry ``predecessors[i, j]`` gives the index of the previous node in the path from point ``i`` to point
        ``j``. If no path exists between nodes ``i`` and ``j``, then ``predecessors[i, j] = -9999``.

    Examples
    --------
    >>> from sknetwork.data import cyclic_digraph
    >>> adjacency = cyclic_digraph(3)
    >>> distance(adjacency, sources=0)
    array([0., 1., 2.])
    >>> distance(adjacency, sources=0, return_predecessors=True)
    (array([0., 1., 2.]), array([-9999,     0,     1]))
    """
    n_jobs = check_n_jobs(n_jobs)
    if method == 'FW' and n_jobs != 1:
        raise ValueError('The Floyd-Warshall algorithm cannot be used with parallel computations.')
    if sources is None:
        sources = np.arange(adjacency.shape[0])
    elif np.issubdtype(type(sources), np.integer):
        sources = np.array([sources])
    n = len(sources)
    directed = not is_symmetric(adjacency)
    local_function = partial(sparse.csgraph.shortest_path,
                             adjacency, method, directed, return_predecessors, unweighted, False)
    if n_jobs == 1 or n == 1:
        res = sparse.csgraph.shortest_path(adjacency, method, directed, return_predecessors,
                                           unweighted, False, sources)
    else:
        with Pool(n_jobs) as pool:
            res = np.array(pool.map(local_function, sources))
    if return_predecessors:
        if n == 1:
            return res[0].ravel(), res[1].astype(int).ravel()
        else:
            return res[0], res[1].astype(int)
    else:
        if n == 1:
            return res.ravel()
        else:
            return res


def shortest_path(adjacency: sparse.csr_matrix, sources: Union[int, Iterable], targets: Union[int, Iterable],
                  method: str = 'D', unweighted: bool = False, n_jobs: Optional[int] = None):
    """Compute the shortest paths in the graph.

    * Graphs
    * Digraphs

    Parameters
    ----------
    adjacency :
        The adjacency matrix of the graph
    sources : int or iterable
        Sources nodes.
    targets : int or iterable
        Target nodes.
    method :
        The method to be used.

        * ``'D'`` (Dijkstra),
        * ``'BF'`` (Bellman-Ford),
        * ``'J'`` (Johnson).
    unweighted :
        If ``True``, the weights of the edges are ignored
    n_jobs :
        If an integer value is given, denotes the number of workers to use (-1 means the maximum number will be used).
        If ``None``, no parallel computations are made.

    Returns
    -------
    paths : list
        If single source and single target, return a list containing the nodes on the path from source to target.
        If multiple sources or multiple targets, return a list of paths as lists.
        An empty list means that the path does not exist.

    Examples
    --------
    >>> from sknetwork.data import linear_digraph
    >>> adjacency = linear_digraph(3)
    >>> shortest_path(adjacency, 0, 2)
    [0, 1, 2]
    >>> shortest_path(adjacency, 2, 0)
    []
    >>> shortest_path(adjacency, 0, [1, 2])
    [[0, 1], [0, 1, 2]]
    >>> shortest_path(adjacency, [0, 1], 2)
    [[0, 1, 2], [1, 2]]
    """
    if np.issubdtype(type(sources), np.integer):
        sources = [sources]
    if np.issubdtype(type(targets), np.integer):
        targets = [targets]

    if len(sources) == 1:
        source2target = True
        source = sources[0]
    elif len(targets) == 1:
        source2target = False
        source = targets[0]
        targets = sources
    else:
        raise ValueError(
            'This request is ambiguous. Either use one source and multiple targets or multiple sources and one target.')

    if source2target:
        dists, preds = distance(adjacency, source, method, True, unweighted, n_jobs)
    else:
        dists, preds = distance(adjacency.T, source, method, True, unweighted, n_jobs)

    paths = []
    for target in targets:
        if dists[target] == np.inf:
            path = []
        else:
            path = [target]
            node = target
            while node != source:
                node = preds[node]
                path.append(node)
        if source2target:
            path.reverse()
        paths.append(path)
    if len(paths) == 1:
        paths = paths[0]
    return paths



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May, 2020
Nathan de Lara <ndelara@enst.fr>
"""
import numpy as np
from scipy import sparse

from sknetwork.utils.format import directed2undirected


def edgelist2adjacency(edgelist: list, undirected: bool = False) -> sparse.csr_matrix:
    """Build an adjacency matrix from a list of edges.

    Parameters
    ----------
    edgelist : list
        List of edges as pairs (i, j) or triplets (i, j, w) for weighted edges.
    undirected : bool
        If ``True``, return a symmetric adjacency.

    Returns
    -------
    adjacency : sparse.csr_matrix

    Examples
    --------
    >>> edgelist = [(0, 1), (1, 2), (2, 0)]
    >>> adjacency = edgelist2adjacency(edgelist)
    >>> adjacency.shape, adjacency.nnz
    ((3, 3), 3)
    >>> adjacency = edgelist2adjacency(edgelist, undirected=True)
    >>> adjacency.shape, adjacency.nnz
    ((3, 3), 6)
    >>> weighted_edgelist = [(0, 1, 0.2), (1, 2, 4), (2, 0, 1.3)]
    >>> adjacency = edgelist2adjacency(weighted_edgelist)
    >>> adjacency.dtype
    dtype('float64')
    """
    edges = np.array(edgelist)
    row, col = edges[:, 0].astype(np.int32), edges[:, 1].astype(np.int32)
    n = max(row.max(), col.max()) + 1
    if edges.shape[1] > 2:
        data = edges[:, 2]
    else:
        data = np.ones_like(row, dtype=bool)
    adjacency = sparse.csr_matrix((data, (row, col)), shape=(n, n))
    if undirected:
        adjacency = directed2undirected(adjacency)
    return adjacency


def edgelist2biadjacency(edgelist: list) -> sparse.csr_matrix:
    """Build a biadjacency matrix from a list of edges.

    Parameters
    ----------
    edgelist : list
        List of edges as pairs (i, j) or triplets (i, j, w) for weighted edges.

    Returns
    -------
    biadjacency : sparse.csr_matrix

    Examples
    --------
    >>> edgelist = [(0, 0), (1, 0), (1, 1), (2, 1)]
    >>> biadjacency = edgelist2biadjacency(edgelist)
    >>> biadjacency.shape, biadjacency.nnz
    ((3, 2), 4)
    >>> weighted_edgelist = [(0, 0, 0.5), (1, 0, 1), (1, 1, 1), (2, 1, 2)]
    >>> biadjacency = edgelist2biadjacency(weighted_edgelist)
    >>> biadjacency.dtype
    dtype('float64')
    """
    edges = np.array(edgelist)
    row, col = edges[:, 0].astype(np.int32), edges[:, 1].astype(np.int32)
    n_row, n_col = row.max() + 1, col.max() + 1
    if edges.shape[1] > 2:
        data = edges[:, 2]
    else:
        data = np.ones_like(row, dtype=bool)
    biadjacency = sparse.csr_matrix((data, (row, col)), shape=(n_row, n_col))
    return biadjacency



#!/usr/bin/env python3
# coding: utf-8
"""
Created on July 9 2019
@author: Nathan De Lara <ndelara@enst.fr>
"""
from abc import ABC
from typing import Union

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import eigsh

from sknetwork.linalg.randomized_methods import randomized_eig
from sknetwork.linalg.sparse_lowrank import SparseLR
from sknetwork.utils.base import Algorithm
from sknetwork.utils.check import check_random_state


class EigSolver(Algorithm, ABC):
    """Generic class for eigensolvers.

    Parameters
    ----------
    which: str
        Which eigenvectors and eigenvalues to find:

        * ``'LM'`` : Largest (in magnitude) eigenvalues.
        * ``'SM'` : Smallest (in magnitude) eigenvalues.

    Attributes
    ----------
    eigenvectors_: np.ndarray
        Two-dimensional array, each column is an eigenvector of the input.
    eigenvalues_: np.ndarray
        Eigenvalues associated to each eigenvector.
    """
    def __init__(self, which='LM'):
        self.which = which

        self.eigenvectors_ = None
        self.eigenvalues_ = None


class LanczosEig(EigSolver):
    """Eigenvalue solver using Lanczos method.

    Parameters
    ----------
    which : str
        Which eigenvectors and eigenvalues to find:

        * ``'LM'`` : Largest (in magnitude) eigenvalues.
        * ``'SM'`` : Smallest (in magnitude) eigenvalues.
    maxiter : int
        Maximum number of Arnoldi update iterations allowed. Default: n*10.
    tol : float
        Relative accuracy for eigenvalues (stopping criterion). The default value of 0 implies machine precision.
    Attributes
    ----------
    eigenvectors_: np.ndarray
        Two-dimensional array, each column is an eigenvector of the input.
    eigenvalues_: np.ndarray
        Eigenvalues associated to each eigenvector.

    See Also
    --------
    scipy.sparse.linalg.eigsh
    """
    def __init__(self, which='LM', maxiter: int = None, tol: float = 0.):
        super(LanczosEig, self).__init__(which=which)
        self.maxiter = maxiter
        self.tol = tol

    def fit(self, matrix: Union[sparse.csr_matrix, sparse.linalg.LinearOperator], n_components: int,
            v0: np.ndarray = None):
        """Perform eigenvalue decomposition on symmetric input matrix.

        Parameters
        ----------
        matrix :
            Matrix to decompose.
        n_components : int
            Number of eigenvectors to compute
        v0 : np.ndarray
            Starting vector for iteration. Default: random.
        Returns
        -------
        self: :class:`EigSolver`
        """
        eigenvalues, eigenvectors = eigsh(matrix.astype(np.float), n_components, which=self.which, maxiter=self.maxiter,
                                          tol=self.tol, v0=v0)
        self.eigenvectors_ = eigenvectors
        self.eigenvalues_ = eigenvalues

        if self.which in ['LM', 'LA']:
            index = np.argsort(self.eigenvalues_)[::-1]
            self.eigenvalues_ = self.eigenvalues_[index]
            self.eigenvectors_ = self.eigenvectors_[:, index]

        return self


class HalkoEig(EigSolver):
    """Eigenvalue solver using Halko's randomized method.

    Parameters
    ----------
    which: str
        Which eigenvectors and eigenvalues to find:

        * ``'LM'`` : Largest (in magnitude) eigenvalues.
        * ``'SM'`` : Smallest (in magnitude) eigenvalues.
    n_oversamples : int (default=10)
        Additional number of random vectors to sample the range of ``matrix`` so as
        to ensure proper conditioning. The total number of random vectors
        used to find the range of ``matrix`` is ``n_components + n_oversamples``. Smaller number can improve speed
        but can negatively impact the quality of approximation of singular vectors and singular values.
    n_iter: int or 'auto' (default is 'auto')
        See :meth:`randomized_range_finder`
    power_iteration_normalizer: ``'auto'`` (default), ``'QR'``, ``'LU'``, ``None``
        See :meth:`randomized_range_finder`
    random_state: int, RandomState instance or None, optional (default=None)
        See :meth:`randomized_range_finder`
    one_pass: bool (default=False)
        whether to use algorithm 5.6 instead of 5.3. 5.6 requires less access to the original matrix,
        while 5.3 is more accurate.
   """
    def __init__(self, which='LM', n_oversamples: int = 10, n_iter='auto',
                 power_iteration_normalizer: Union[str, None] = 'auto', random_state=None, one_pass: bool = False):
        super(HalkoEig, self).__init__(which=which)
        self.n_oversamples = n_oversamples
        self.n_iter = n_iter
        self.power_iteration_normalizer = power_iteration_normalizer
        self.random_state = check_random_state(random_state)
        self.one_pass = one_pass

    def fit(self, matrix: Union[sparse.csr_matrix, sparse.linalg.LinearOperator, SparseLR], n_components: int):
        """Perform eigenvalue decomposition on input matrix.

        Parameters
        ----------
        matrix :
            Matrix to decompose.
        n_components : int
            Number of eigenvectors to compute

        Returns
        -------
        self: :class:`EigSolver`
        """
        eigenvalues, eigenvectors = randomized_eig(matrix, n_components, self.which, self.n_oversamples, self.n_iter,
                                                   self.power_iteration_normalizer, self.random_state, self.one_pass)
        self.eigenvectors_ = eigenvectors
        self.eigenvalues_ = eigenvalues

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for hierarchy API"""
import unittest

from sknetwork.data.test_graphs import *
from sknetwork.embedding import GSVD
from sknetwork.hierarchy import *


class TestHierarchyAPI(unittest.TestCase):

    def test_undirected(self):
        adjacency = test_graph()
        n = adjacency.shape[0]

        hierarchy = [Paris(), Ward(GSVD(3)), LouvainHierarchy()]
        for hier_algo in hierarchy:
            dendrogram = hier_algo.fit_transform(adjacency)
            self.assertTupleEqual(dendrogram.shape, (n - 1, 4))

    def test_disconnected(self):
        adjacency = test_graph_disconnect()
        hierarchy = [Paris(), Ward(GSVD(3)), LouvainHierarchy()]
        for hier_algo in hierarchy:
            dendrogram = hier_algo.fit_transform(adjacency)
            self.assertEqual(dendrogram.shape, (9, 4))

    def test_bipartite(self):
        biadjacency = test_bigraph()
        n_row, n_col = biadjacency.shape

        hierarchy = [BiParis(), BiWard(GSVD(3), cluster_col=True, cluster_both=True), BiLouvainHierarchy()]
        for algo in hierarchy:
            algo.fit_transform(biadjacency)
            self.assertTupleEqual(algo.dendrogram_row_.shape, (n_row - 1, 4))
            self.assertTupleEqual(algo.dendrogram_col_.shape, (n_col - 1, 4))
            if algo.dendrogram_full_ is not None:
                self.assertTupleEqual(algo.dendrogram_full_.shape, (n_row + n_col - 1, 4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for label propagation"""

import unittest

from sknetwork.classification import Propagation
from sknetwork.data import karate_club


class TestLabelPropagation(unittest.TestCase):

    def test_options(self):
        adjacency = karate_club(metadata=False)
        n = adjacency.shape[0]
        seeds = {0: 0, 1: 1}
        propagation = Propagation(n_iter=3, weighted=False)
        labels = propagation.fit_transform(adjacency, seeds)
        self.assertEqual(labels.shape, (n,))

        for order in ['random', 'decreasing', 'increasing']:
            propagation = Propagation(node_order=order)
            labels = propagation.fit_transform(adjacency, seeds)
            self.assertEqual(labels.shape, (n,))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for link prediction postprocessing"""
import unittest

import numpy as np

from sknetwork.data import house
from sknetwork.linkpred import is_edge


class TestLinkPredPostProcessing(unittest.TestCase):

    def test_signature(self):
        adjacency = house()

        with self.assertRaises(ValueError):
            is_edge(adjacency, 'toto')

        with self.assertRaises(ValueError):
            is_edge(adjacency, np.ones(8).reshape((2, 2, 2)))



#!/usr/bin/env python3
import contextlib
import signal


class TimeOut(contextlib.ContextDecorator):
    """
    Timeout context manager/decorator.

    Adapted from https://gist.github.com/TySkby/143190ad1b88c6115597c45f996b030c on 12/10/2020.

    Examples
    --------
    >>> from time import sleep
    >>> try:
    ...     with TimeOut(1):
    ...         sleep(10)
    ... except TimeoutError:
    ...     print("Function timed out")
    Function timed out
    """
    def __init__(self, seconds: float):
        self.seconds = seconds

    def _timeout_handler(self, signum, frame):
        raise TimeoutError("Code timed out.")

    def __enter__(self):
        signal.signal(signal.SIGALRM, self._timeout_handler)
        signal.alarm(self.seconds)

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for clustering API"""
import unittest

from sknetwork.clustering import *
from sknetwork.data import house, star_wars
from sknetwork.data.test_graphs import *
from sknetwork.embedding.svd import GSVD


class TestClusteringAPI(unittest.TestCase):

    def test_regular(self):

        clustering = [Louvain(return_aggregate=True), KMeans(embedding_method=GSVD(3), return_aggregate=True),
                      PropagationClustering(return_aggregate=True)]
        for clustering_algo in clustering:
            for adjacency in [test_graph(), test_digraph(), test_graph_disconnect()]:
                n = adjacency.shape[0]
                adjacency_bool = adjacency.copy()
                adjacency_bool.data = adjacency_bool.data.astype(bool)

                labels1 = clustering_algo.fit_transform(adjacency)
                labels2 = clustering_algo.fit_transform(adjacency_bool)
                self.assertEqual(labels1.shape, (n,))
                self.assertEqual(labels2.shape, (n,))

                n_labels = len(set(labels2))
                self.assertEqual(clustering_algo.adjacency_.shape, (n_labels, n_labels))

    def test_bipartite(self):
        biadjacency = test_bigraph()
        n_row, n_col = biadjacency.shape

        clustering = [BiLouvain(return_aggregate=True),
                      BiKMeans(embedding_method=GSVD(3), co_cluster=True, return_aggregate=True),
                      BiPropagationClustering(return_aggregate=True)]
        for clustering_algo in clustering:
            clustering_algo.fit_transform(biadjacency)
            labels_row = clustering_algo.labels_row_
            labels_col = clustering_algo.labels_col_
            self.assertEqual(labels_row.shape, (n_row,))
            self.assertEqual(labels_col.shape, (n_col,))

    def test_aggregate(self):
        adjacency = house()
        biadjacency = star_wars()

        louvain = Louvain(return_aggregate=True)
        louvain.fit(adjacency)
        self.assertTrue(np.issubdtype(louvain.adjacency_.dtype, np.float_))

        bilouvain = BiLouvain(return_aggregate=True)
        bilouvain.fit(biadjacency)
        self.assertTrue(np.issubdtype(bilouvain.biadjacency_.dtype, np.float_))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for diffusion.py"""

import unittest

from sknetwork.data.test_graphs import *
from sknetwork.ranking import Diffusion, BiDiffusion, Dirichlet, BiDirichlet


# noinspection DuplicatedCode
class TestDiffusion(unittest.TestCase):

    def setUp(self):
        self.algos = [Diffusion(), Dirichlet()]

    def test_n_iter(self):
        with self.assertRaises(ValueError):
            Diffusion(n_iter=-1)

    def test_single_node_graph(self):
        for algo in self.algos:
            algo.fit(sparse.identity(1, format='csr'), {0: 1})
            self.assertEqual(algo.scores_, [1])

    def test_range(self):
        for adjacency in [test_graph(), test_digraph()]:
            for algo in self.algos:
                scores = algo.fit_transform(adjacency, {0: 0, 1: 1, 2: 0.5})
                self.assertTrue(np.all(scores <= 1) and np.all(scores >= 0))

        biadjacency = test_bigraph()
        for algo in [BiDiffusion(), BiDirichlet()]:
            scores = algo.fit_transform(biadjacency, {0: 1})
            self.assertTrue(np.all(scores <= 1) and np.all(scores >= 0))
            scores = algo.fit_transform(biadjacency, {0: 0.1}, {1: 2}, 0.3)
            self.assertTrue(np.all(scores <= 2) and np.all(scores >= 0.1))

    def test_initial_state(self):
        for adjacency in [test_graph(), test_digraph()]:
            for algo in self.algos:
                scores = algo.fit_transform(adjacency, {0: 0, 1: 1, 2: 0.5}, 0.3)
                self.assertTrue(np.all(scores <= 1) and np.all(scores >= 0))

    def test_damping_factor(self):
        for adjacency in [test_graph(), test_digraph()]:
            for algo in [Diffusion(damping_factor=0.5), Dirichlet(damping_factor=0.5)]:
                scores = algo.fit_transform(adjacency, {0: 0, 1: 1, 2: 0.5})
                self.assertTrue(np.all(scores <= 1) and np.all(scores >= 0))

    def test_bicgstab(self):
        algo1 = Dirichlet(n_iter=0)
        algo2 = Dirichlet(n_iter=10)
        adjacency = test_digraph()
        scores1 = algo1.fit_transform(adjacency, {0: 1})
        scores2 = algo2.fit_transform(adjacency, {0: 1})
        self.assertAlmostEqual(np.linalg.norm(scores1 - scores2), 0)
        algo1 = Dirichlet(n_iter=0, damping_factor=0.5)
        algo2 = Dirichlet(n_iter=10, damping_factor=0.5)
        adjacency = test_digraph()
        scores1 = algo1.fit_transform(adjacency, {0: 1})
        scores2 = algo2.fit_transform(adjacency, {0: 1})
        self.assertAlmostEqual(np.linalg.norm(scores1 - scores2), 0)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for embeddings metrics"""

import unittest

from sknetwork.data.test_graphs import test_bigraph, test_graph
from sknetwork.embedding import GSVD, LouvainEmbedding
from sknetwork.embedding.metrics import cosine_modularity


class TestClusteringMetrics(unittest.TestCase):

    def test_cosine(self):
        biadjacency = test_bigraph()
        method = GSVD(3, solver='lanczos')

        embedding = method.fit_transform(biadjacency)
        embedding_col = method.embedding_col_
        fit, div, modularity = cosine_modularity(biadjacency, embedding, embedding_col, return_all=True)
        modularity = cosine_modularity(biadjacency, embedding, embedding_col, return_all=False)
        self.assertAlmostEqual(modularity, fit - div)

        adjacency = test_graph()
        embedding = method.fit_transform(adjacency)
        fit, div, modularity = cosine_modularity(adjacency, embedding, return_all=True)
        self.assertAlmostEqual(modularity, fit - div)

        with self.assertRaises(ValueError):
            cosine_modularity(biadjacency, embedding)

        louvain = LouvainEmbedding()
        embedding = louvain.fit_transform(adjacency)
        fit, div, modularity = cosine_modularity(adjacency, embedding, return_all=True)
        self.assertAlmostEqual(modularity, fit - div)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu July 10 2018
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <bonald@enst.fr>
"""
from typing import Union, Tuple

import numpy as np
from scipy import sparse

from sknetwork.linalg.normalization import diag_pinv
from sknetwork.utils.check import check_format, check_probs, check_square
from sknetwork.utils.format import bipartite2directed
from sknetwork.utils.membership import membership_matrix


def modularity(adjacency: Union[sparse.csr_matrix, np.ndarray], labels: np.ndarray,
               weights: Union[str, np.ndarray] = 'degree', weights_in: Union[str, np.ndarray] = 'degree',
               resolution: float = 1, return_all: bool = False) -> Union[float, Tuple[float, float, float]]:
    """Modularity of a clustering (node partition).

    * Graphs
    * Digraphs

    The modularity of a clustering is

    :math:`Q = \\dfrac{1}{w} \\sum_{i,j}\\left(A_{ij} - \\gamma \\dfrac{d_id_j}{w}\\right)\\delta_{c_i,c_j}`
    for graphs,

    :math:`Q = \\dfrac{1}{w} \\sum_{i,j}\\left(A_{ij} - \\gamma \\dfrac{d^+_id^-_j}{w}\\right)\\delta_{c_i,c_j}`
    for digraphs,

    where

    * :math:`c_i` is the cluster of node :math:`i`,\n
    * :math:`d_i` is the weight of node :math:`i`,\n
    * :math:`d^+_i, d^-_i` are the out-weight, in-weight of node :math:`i` (for digraphs),\n
    * :math:`w = 1^TA1` is the total weight,\n
    * :math:`\\delta` is the Kronecker symbol,\n
    * :math:`\\gamma \\ge 0` is the resolution parameter.

    Parameters
    ----------
    adjacency:
        Adjacency matrix of the graph.
    labels:
        Labels of nodes, vector of size :math:`n` .
    weights :
        Weights of nodes.
        ``'degree'`` (default), ``'uniform'`` or custom weights.
    weights_in :
        In-weights of nodes.
        ``None`` (default), ``'degree'``, ``'uniform'`` or custom weights.
        If ``None``, taken equal to weights.
    resolution:
        Resolution parameter (default = 1).
    return_all:
        If ``True``, return modularity, fit, diversity.

    Returns
    -------
    modularity : float
    fit: float, optional
    diversity: float, optional

    Example
    -------
    >>> from sknetwork.clustering import modularity
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> labels = np.array([0, 0, 1, 1, 0])
    >>> np.round(modularity(adjacency, labels), 2)
    0.11
    """
    adjacency = check_format(adjacency).astype(float)
    check_square(adjacency)

    if len(labels) != adjacency.shape[0]:
        raise ValueError('Dimension mismatch between labels and adjacency matrix.')

    probs_row = check_probs(weights, adjacency)
    probs_col = check_probs(weights_in, adjacency.T)
    membership = membership_matrix(labels)

    fit: float = membership.multiply(adjacency.dot(membership)).data.sum() / adjacency.data.sum()
    div: float = membership.T.dot(probs_col).dot(membership.T.dot(probs_row))
    mod: float = fit - resolution * div
    if return_all:
        return mod, fit, div
    else:
        return mod


def bimodularity(biadjacency: Union[sparse.csr_matrix, np.ndarray], labels: np.ndarray, labels_col: np.ndarray,
                 weights: Union[str, np.ndarray] = 'degree', weights_col: Union[str, np.ndarray] = 'degree',
                 resolution: float = 1, return_all: bool = False) -> Union[float, Tuple[float, float, float]]:
    """Bimodularity of a clustering (node partition).

    * Bigraphs

    The bimodularity of a clustering is

    :math:`Q = \\sum_{i}\\sum_{j}\\left(\\dfrac{B_{ij}}{w} - \\gamma \\dfrac{d_{1,i}d_{2,j}}{w^2}\\right)
    \\delta_{c_{1,i},c_{2,j}}`

    where

    * :math:`c_{1,i}, c_{2,j}` are the clusters of nodes :math:`i` (row) and :math:`j` (column),\n
    * :math:`d_{1,i}, d_{2,j}` are the weights of nodes :math:`i` (row) and :math:`j` (column),\n
    * :math:`w = 1^TB1` is the total weight,\n
    * :math:`\\delta` is the Kronecker symbol,\n
    * :math:`\\gamma \\ge 0` is the resolution parameter.

    Parameters
    ----------
    biadjacency :
        Biadjacency matrix of the graph (shape :math:`n_1 \\times n_2`).
    labels :
        Labels of rows, vector of size :math:`n_1`.
    labels_col:
        Labels of columns, vector of size :math:`n_2`.
    weights :
        Weights of nodes.
        ``'degree'`` (default), ``'uniform'`` or custom weights.
    weights_col :
        Weights of columns.
        ``'degree'`` (default), ``'uniform'`` or custom weights.
    resolution:
        Resolution parameter (default = 1).
    return_all:
        If ``True``, return modularity, fit, diversity.

    Returns
    -------
    modularity : float
    fit: float, optional
    diversity: float, optional

    Example
    -------
    >>> from sknetwork.clustering import bimodularity
    >>> from sknetwork.data import star_wars
    >>> biadjacency = star_wars()
    >>> labels = np.array([1, 1, 0, 0])
    >>> labels_col = np.array([1, 0, 0])
    >>> np.round(bimodularity(biadjacency, labels, labels_col), 2)
    0.22
    """
    biadjacency = check_format(biadjacency).astype(float)
    n_row, n_col = biadjacency.shape

    if len(labels) != n_row:
        raise ValueError('Dimension mismatch between labels and biadjacency matrix.')
    if len(labels_col) != n_col:
        raise ValueError('Dimension mismatch between labels_col and biadjacency matrix.')

    adjacency = bipartite2directed(biadjacency)

    weights_ = check_probs(weights, biadjacency)
    weights_ = np.hstack((weights_, np.zeros(n_col)))
    weights_col_ = check_probs(weights_col, biadjacency.T)
    weights_col_ = np.hstack((np.zeros(n_row), weights_col_))

    labels_ = np.hstack((labels, labels_col))

    return modularity(adjacency, labels_, weights_, weights_col_, resolution, return_all)


def comodularity(adjacency: Union[sparse.csr_matrix, np.ndarray], labels: np.ndarray, resolution: float = 1,
                 return_all: bool = False) -> Union[float, Tuple[float, float, float]]:
    """Modularity of a clustering in the normalized co-neighborhood graph.

    * Graphs
    * Digraphs
    * Bigraphs

    Quality metric of a clustering given by:

    :math:`Q = \\dfrac{1}{w}\\sum_{i,j}\\left((AD_2^{-1}A^T)_{ij} - \\gamma \\dfrac{d_id_j}{w}\\right)
    \\delta_{c_i,c_j}`

    where

    * :math:`c_i` is the cluster of node `i`,\n
    * :math:`\\delta` is the Kronecker symbol,\n
    * :math:`\\gamma \\ge 0` is the resolution parameter.

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    labels :
       Labels of the nodes.
    resolution :
        Resolution parameter (default = 1).
    return_all :
        If ``True``, return modularity, fit, diversity.

    Returns
    -------
    modularity : float
    fit : float, optional
    diversity: float, optional

    Example
    -------
    >>> from sknetwork.clustering import comodularity
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> labels = np.array([0, 0, 1, 1, 0])
    >>> np.round(comodularity(adjacency, labels), 2)
    0.06

    Notes
    -----
    Does not require the computation of the adjacency matrix of the normalized co-neighborhood graph.
    """
    adjacency = check_format(adjacency).astype(float)

    n_row, n_col = adjacency.shape
    total_weight = adjacency.data.sum()
    probs = adjacency.dot(np.ones(n_col)) / total_weight

    weights_col = adjacency.T.dot(np.ones(n_col))
    diag_col = diag_pinv(np.sqrt(weights_col))
    normalized_adjacency = (adjacency.dot(diag_col)).T.tocsr()

    if len(labels) != n_row:
        raise ValueError('The number of labels must match the number of rows.')

    membership = membership_matrix(labels)
    fit: float = ((normalized_adjacency.dot(membership)).data ** 2).sum() / total_weight
    div: float = np.linalg.norm(membership.T.dot(probs)) ** 2
    mod: float = fit - resolution * div

    if return_all:
        return mod, fit, div
    else:
        return mod


def normalized_std(labels: np.ndarray) -> float:
    """Normalized standard deviation of cluster sizes.

    A score of 1 means perfectly balanced clustering.

    Parameters
    ----------
    labels :
        Labels of nodes.

    Returns
    -------
    float

    Example
    -------
    >>> from sknetwork.clustering import normalized_std
    >>> labels = np.array([0, 0, 1, 1])
    >>> normalized_std(labels)
    1.0
    """
    n = labels.shape[0]
    unique_clusters, counts = np.unique(labels, return_counts=True)
    n_clust = len(unique_clusters)
    if n_clust < 2:
        raise ValueError('There must be at least two different clusters.')
    return 1 - np.std(counts) / np.sqrt(n ** 2 * (n_clust - 1) / n_clust ** 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""""tests for dag.py"""
import unittest

import numpy as np

from sknetwork.data import house
from sknetwork.topology import DAG


class TestDAG(unittest.TestCase):

    def test_options(self):
        adjacency = house()
        dag = DAG()
        dag.fit(adjacency)
        self.assertEqual(dag.indptr_.shape[0], adjacency.shape[0]+1)
        self.assertEqual(dag.indices_.shape[0], 6)

        with self.assertRaises(ValueError):
            dag.fit(adjacency, sorted_nodes=np.arange(3))

        with self.assertRaises(ValueError):
            dag = DAG(ordering='toto')
            dag.fit(adjacency)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for verbose.py"""

import unittest

from sknetwork.utils.verbose import VerboseMixin


class TestVerbose(unittest.TestCase):

    def test_prints(self):
        verbose = VerboseMixin(verbose=True)
        verbose.log.print('There are', 4, 'seasons in a year')
        self.assertEqual(str(verbose.log), 'There are 4 seasons in a year\n')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Weisfeiler-Lehman coloring"""
import unittest

from sknetwork.data import house, bow_tie
from sknetwork.data.test_graphs import *
from sknetwork.topology import WeisfeilerLehman


class TestWLColoring(unittest.TestCase):

    def test_empty(self):
        adjacency = test_graph_empty()
        labels = WeisfeilerLehman().fit_transform(adjacency)
        self.assertTrue((labels == np.zeros(10)).all())

    def test_cliques(self):
        adjacency = test_graph_clique()
        labels = WeisfeilerLehman().fit_transform(adjacency)
        self.assertTrue((labels == np.zeros(10)).all())

    def test_house(self):
        adjacency = house()
        labels = WeisfeilerLehman().fit_transform(adjacency)
        self.assertTrue((labels == np.array([0, 2, 1, 1, 2])).all())

    def test_bow_tie(self):
        adjacency = bow_tie()
        labels = WeisfeilerLehman().fit_transform(adjacency)
        self.assertTrue((labels == np.array([1, 0, 0, 0, 0])).all())

    def test_iso(self):
        adjacency = house()
        n = adjacency.indptr.shape[0] - 1
        reorder = list(range(n))
        np.random.shuffle(reorder)
        adjacency2 = adjacency[reorder][:, reorder]
        l1 = WeisfeilerLehman().fit_transform(adjacency)
        l2 = WeisfeilerLehman().fit_transform(adjacency2)
        l1.sort()
        l2.sort()
        self.assertTrue((l1 == l2).all())

    def test_early_stop(self):
        adjacency = house()
        wl = WeisfeilerLehman(max_iter=1)
        labels = wl.fit_transform(adjacency)
        self.assertTrue((labels == np.array([0, 1, 0, 0, 1])).all())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on April 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""

import unittest

import numpy as np
from scipy import sparse

from sknetwork.linalg import normalize, CoNeighborOperator


class TestNormalization(unittest.TestCase):

    def test_formats(self):
        n = 5
        mat1 = normalize(np.eye(n))
        mat2 = normalize(sparse.eye(n))
        mat3 = normalize(CoNeighborOperator(mat2))

        x = np.random.randn(n)
        self.assertAlmostEqual(np.linalg.norm(mat1.dot(x) - x), 0)
        self.assertAlmostEqual(np.linalg.norm(mat2.dot(x) - x), 0)
        self.assertAlmostEqual(np.linalg.norm(mat3.dot(x) - x), 0)

        mat1 = np.random.rand(n**2).reshape((n, n))
        mat2 = sparse.csr_matrix(mat1)
        mat1 = normalize(mat1, p=2)
        mat2 = normalize(mat2, p=2)
        self.assertAlmostEqual(np.linalg.norm(mat1.dot(x) - mat2.dot(x)), 0)

        with self.assertRaises(NotImplementedError):
            normalize(mat3, p=2)
        with self.assertRaises(NotImplementedError):
            normalize(mat1, p=3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 19 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Union, Tuple

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import LinearOperator


class SparseLR(LinearOperator):
    """Class for matrices with "sparse + low rank" structure.
    Example:

    :math:`A + xy^T`

    Parameters
    ----------
    sparse_mat: scipy.spmatrix
        Sparse component. Is converted to csr format automatically.
    low_rank_tuples: list
        Single tuple of arrays of list of tuples, representing the low rank components [(x1, y1), (x2, y2),...].
        Each low rank component is of the form :math:`xy^T`.

    Examples
    --------
    >>> from scipy import sparse
    >>> from sknetwork.linalg import SparseLR
    >>> adjacency = sparse.eye(2, format='csr')
    >>> slr = SparseLR(adjacency, (np.ones(2), np.ones(2)))
    >>> x = np.ones(2)
    >>> slr.dot(x)
    array([3., 3.])
    >>> slr.sum(axis=0)
    array([3., 3.])
    >>> slr.sum(axis=1)
    array([3., 3.])
    >>> slr.sum()
    6.0

    References
    ----------
    De Lara (2019). `The Sparse + Low Rank trick for Matrix Factorization-Based Graph Algorithms.
    <http://www.mlgworkshop.org/2019/papers/MLG2019_paper_1.pdf>`_
    Proceedings of the 15th International Workshop on Mining and Learning with Graphs (MLG).
    """
    def __init__(self, sparse_mat: Union[sparse.csr_matrix, sparse.csc_matrix], low_rank_tuples: Union[list, Tuple],
                 dtype=float):
        n_row, n_col = sparse_mat.shape
        self.sparse_mat = sparse_mat.tocsr().astype(dtype)
        super(SparseLR, self).__init__(dtype=dtype, shape=(n_row, n_col))

        if isinstance(low_rank_tuples, Tuple):
            low_rank_tuples = [low_rank_tuples]
        self.low_rank_tuples = []
        for x, y in low_rank_tuples:
            if x.shape == (n_row,) and y.shape == (n_col,):
                self.low_rank_tuples.append((x.astype(self.dtype), y.astype(self.dtype)))
            else:
                raise ValueError('For each low rank tuple, x (resp. y) should be a vector of length {} (resp. {})'
                                 .format(n_row, n_col))

    def __neg__(self):
        return SparseLR(-self.sparse_mat, [(-x, y) for (x, y) in self.low_rank_tuples])

    def __add__(self, other: 'SparseLR'):
        if type(other) == sparse.csr_matrix:
            return SparseLR(self.sparse_mat + other, self.low_rank_tuples)
        else:
            return SparseLR(self.sparse_mat + other.sparse_mat, self.low_rank_tuples + other.low_rank_tuples)

    def __sub__(self, other):
        return self.__add__(-other)

    def __mul__(self, other):
        return SparseLR(other * self.sparse_mat, [(other * x, y) for (x, y) in self.low_rank_tuples])

    def _matvec(self, matrix: np.ndarray):
        """Right dot product with a dense matrix.

        Parameters
        ----------
        matrix:
            Matrix.

        Returns
        -------
        Dot product as a dense array
        """
        prod = self.sparse_mat.dot(matrix)
        if len(matrix.shape) == 1:
            for (x, y) in self.low_rank_tuples:
                prod += x * matrix.dot(y)
        else:
            transposed = matrix.T
            for (x, y) in self.low_rank_tuples:
                prod += x[:, np.newaxis].dot(transposed.dot(y)[:, np.newaxis].T)
        return prod

    def _transpose(self):
        """Transposed operator."""
        transposed_sparse = sparse.csr_matrix(self.sparse_mat.T)
        transposed_tuples = [(y, x) for (x, y) in self.low_rank_tuples]
        return SparseLR(transposed_sparse, transposed_tuples)

    def _adjoint(self):
        return self.transpose()

    def left_sparse_dot(self, matrix: sparse.csr_matrix):
        """Left dot product with a sparse matrix."""
        return SparseLR(matrix.dot(self.sparse_mat), [(matrix.dot(x), y) for (x, y) in self.low_rank_tuples])

    def right_sparse_dot(self, matrix: sparse.csr_matrix):
        """Right dot product with a sparse matrix."""
        return SparseLR(self.sparse_mat.dot(matrix), [(x, matrix.T.dot(y)) for (x, y) in self.low_rank_tuples])

    def sum(self, axis=None):
        """Row-wise, column-wise or total sum of operator's coefficients.

        Parameters
        ----------
        axis :
            If 0, return column-wise sum. If 1, return row-wise sum. Otherwise, return total sum.
        """
        if axis == 0:
            s = self.T.dot(np.ones(self.shape[0]))
        elif axis == 1:
            s = self.dot(np.ones(self.shape[1]))
        else:
            s = self.dot(np.ones(self.shape[1])).sum()
        return s

    def astype(self, dtype: Union[str, np.dtype]):
        """Change dtype of the object."""
        self.sparse_mat = self.sparse_mat.astype(dtype)
        self.low_rank_tuples = [(x.astype(dtype), y.astype(dtype)) for (x, y) in self.low_rank_tuples]
        self.dtype = np.dtype(dtype)

        return self



"""path module"""
from sknetwork.topology.kcliques import Cliques
from sknetwork.topology.kcore import CoreDecomposition
from sknetwork.topology.triangles import Triangles

from sknetwork.topology.dag import DAG
from sknetwork.topology.structure import is_acyclic, is_bipartite, largest_connected_component, connected_components
from sknetwork.topology.weisfeiler_lehman import WeisfeilerLehman, are_isomorphic



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for bunch"""

import unittest

from sknetwork.utils import Bunch


class TestBunch(unittest.TestCase):

    def test_key_error(self):
        bunch = Bunch(a=1, b=2)
        with self.assertRaises(AttributeError):
            # noinspection PyStatementEffect
            bunch.c



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""

from typing import Union

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import LinearOperator

from sknetwork.utils.check import check_format, check_square


class Polynome(LinearOperator):
    """Polynome of an adjacency matrix as a linear operator

    :math:`P(A) = \\alpha_k A^k + ... + \\alpha_1 A + \\alpha_0`.

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph
    coeffs : np.ndarray
        Coefficients of the polynome by increasing order of power.

    Examples
    --------
    >>> from scipy import sparse
    >>> from sknetwork.linalg import Polynome
    >>> adjacency = sparse.eye(2, format='csr')
    >>> polynome = Polynome(adjacency, np.arange(3))
    >>> x = np.ones(2)
    >>> polynome.dot(x)
    array([3., 3.])
    >>> polynome.T.dot(x)
    array([3., 3.])

    Notes
    -----
    The polynome is evaluated using the `Ruffini-Horner method
    <https://en.wikipedia.org/wiki/Horner%27s_method>`_.
    """

    def __init__(self, adjacency: Union[sparse.csr_matrix, np.ndarray], coeffs: np.ndarray):
        if coeffs.shape[0] == 0:
            raise ValueError('A polynome requires at least one coefficient.')
        if not isinstance(adjacency, LinearOperator):
            adjacency = check_format(adjacency)
        check_square(adjacency)
        shape = adjacency.shape
        dtype = adjacency.dtype
        super(Polynome, self).__init__(dtype=dtype, shape=shape)

        self.adjacency = adjacency
        self.coeffs = coeffs

    def __neg__(self):
        return Polynome(self.adjacency, -self.coeffs)

    def __mul__(self, other):
        return Polynome(self.adjacency, other * self.coeffs)

    def _matvec(self, matrix: np.ndarray):
        """Right dot product with a dense matrix.
        """
        y = self.coeffs[-1] * matrix
        for a in self.coeffs[::-1][1:]:
            y = self.adjacency.dot(y) + a * matrix
        return y

    def _transpose(self):
        """Transposed operator."""
        return Polynome(self.adjacency.T.tocsr(), self.coeffs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
import numpy as np
from scipy.cluster.vq import kmeans2

from sknetwork.utils.base import Algorithm


class KMeansDense(Algorithm):
    """Standard KMeansDense clustering based on SciPy function ``kmeans2``.

    Parameters
    ----------
    n_clusters :
        Number of desired clusters.
    init :
        Method for initialization. Available methods are ‘random’, ‘points’, ‘++’ and ‘matrix’:
        * ‘random’: generate k centroids from a Gaussian with mean and variance estimated from the data.
        * ‘points’: choose k observations (rows) at random from data for the initial centroids.
        * ‘++’: choose k observations accordingly to the kmeans++ method (careful seeding)
        * ‘matrix’: interpret the k parameter as a k by M (or length k array for one-dimensional data) array of initial
        centroids.
    n_init :
        Number of iterations of the k-means algorithm to run.
    tol :
        Relative tolerance with regards to inertia to declare convergence.

    Attributes
    ----------
    labels_ :
        Label of each sample.
    cluster_centers_ :
        A ‘k’ by ‘N’ array of centroids found at the last iteration of k-means.

    References
    ----------
    * MacQueen, J. (1967, June). Some methods for classification and analysis of multivariate observations.
      In Proceedings of the fifth Berkeley symposium on mathematical statistics and probability
      (Vol. 1, No. 14, pp. 281-297).

    * Arthur, D., & Vassilvitskii, S. (2007, January). k-means++: The advantages of careful seeding.
      In Proceedings of the eighteenth annual ACM-SIAM symposium on Discrete algorithms (pp. 1027-1035).
      Society for Industrial and Applied Mathematics.
    """
    def __init__(self, n_clusters: int = 8, init: str = '++', n_init: int = 10, tol: float = 1e-4):
        self.n_clusters = n_clusters
        self.init = init.lower()
        self.n_init = n_init
        self.tol = tol

        self.labels_ = None
        self.cluster_centers_ = None

    def fit(self, x: np.ndarray) -> 'KMeansDense':
        """Fit algorithm to the data.

        Parameters
        ----------
        x:
            Data to cluster.

        Returns
        -------
        self: :class:`KMeansDense`
        """
        centroids, labels = kmeans2(data=x, k=self.n_clusters, iter=self.n_init, thresh=self.tol, minit=self.init)
        self.cluster_centers_ = centroids
        self.labels_ = labels

        return self

    def fit_transform(self, x: np.ndarray) -> np.ndarray:
        """Fit algorithm to the data and return the labels.

        Parameters
        ----------
        x:
            Data to cluster.

        Returns
        -------
        labels: np.ndarray
        """
        self.fit(x)
        return self.labels_



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
import unittest

import numpy as np

from sknetwork.data import movie_actor
from sknetwork.linalg import CoNeighborOperator
from sknetwork.utils import co_neighbor_graph


class TestCoNeighbors(unittest.TestCase):

    def setUp(self):
        """Simple biadjacency for tests."""
        self.biadjacency = movie_actor()

    def test_exact(self):
        n = self.biadjacency.shape[0]
        adjacency = co_neighbor_graph(self.biadjacency, method='exact', normalized=False)
        self.assertEqual(adjacency.shape, (n, n))
        adjacency = co_neighbor_graph(self.biadjacency, method='exact')
        self.assertEqual(adjacency.shape, (n, n))

        operator = CoNeighborOperator(self.biadjacency)
        x = np.random.randn(n)
        y1 = adjacency.dot(x)
        y2 = operator.dot(x)
        self.assertAlmostEqual(np.linalg.norm(y1 - y2), 0)

    def test_knn(self):
        n = self.biadjacency.shape[0]
        adjacency = co_neighbor_graph(self.biadjacency, method='knn')
        self.assertEqual(adjacency.shape, (n, n))
        adjacency = co_neighbor_graph(self.biadjacency, method='knn', normalized=False)
        self.assertEqual(adjacency.shape, (n, n))

    def test_invalid(self):
        self.assertRaises(ValueError, co_neighbor_graph, self.biadjacency, method='toto')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
import unittest

import numpy as np

from sknetwork.utils import WardDense


class TestKMeans(unittest.TestCase):

    def test_kmeans(self):
        x = np.random.randn(10, 3)
        ward = WardDense()
        dendrogram = ward.fit_transform(x)
        self.assertEqual(dendrogram.shape, (x.shape[0] - 1, 4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov, 2019
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <tbonald@enst.fr>
"""
from typing import Optional, Union

import numpy as np
from scipy import sparse
from scipy.spatial import cKDTree

from sknetwork.classification.base import BaseClassifier, BaseBiClassifier
from sknetwork.embedding.base import BaseEmbedding
from sknetwork.embedding.svd import GSVD
from sknetwork.linalg.normalization import normalize
from sknetwork.utils.check import check_seeds, check_n_neighbors, check_n_jobs
from sknetwork.utils.seeds import stack_seeds


class KNN(BaseClassifier):
    """Node classification by K-nearest neighbors in the embedding space.

    * Graphs
    * Digraphs
    * Bigraphs

    For bigraphs, classify rows only (see ``BiKNN`` for joint classification of rows and columns).

    Parameters
    ----------
    embedding_method :
        Which algorithm to use to project the nodes in vector space. Default is ``GSVD``.
    n_neighbors :
        Number of nearest neighbors to consider.
    factor_distance :
        Power weighting factor :math:`\\alpha` applied to the distance to each neighbor.
        Neighbor at distance :math:``d`` has weight :math:`1 / d^\\alpha`. Default is 2.
    leaf_size :
        Leaf size passed to KDTree.
    p :
        Which Minkowski p-norm to use. Default is 2 (Euclidean distance).
    tol_nn :
        Tolerance in nearest neighbors search; the k-th returned value is guaranteed to be no further
        than ``1 + tol_nn`` times the distance to the actual k-th nearest neighbor.
    n_jobs :
        Number of jobs to schedule for parallel processing. If -1 is given all processors are used.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node.
    membership_ : sparse.csr_matrix
        Membership matrix (columns = labels).

    Example
    -------
    >>> from sknetwork.classification import KNN
    >>> from sknetwork.embedding import GSVD
    >>> from sknetwork.data import karate_club
    >>> knn = KNN(GSVD(3), n_neighbors=1)
    >>> graph = karate_club(metadata=True)
    >>> adjacency = graph.adjacency
    >>> labels_true = graph.labels
    >>> seeds = {0: labels_true[0], 33: labels_true[33]}
    >>> labels_pred = knn.fit_transform(adjacency, seeds)
    >>> np.round(np.mean(labels_pred == labels_true), 2)
    0.97
    """
    def __init__(self, embedding_method: BaseEmbedding = GSVD(10), n_neighbors: int = 5,
                 factor_distance: float = 2, leaf_size: int = 16, p: float = 2, tol_nn: float = 0.01,
                 n_jobs: Optional[int] = None):
        super(KNN, self).__init__()

        self.embedding_method = embedding_method
        self.n_neighbors = n_neighbors
        self.factor_distance = factor_distance
        self.leaf_size = leaf_size
        self.p = p
        self.tol_nn = tol_nn
        self.n_jobs = check_n_jobs(n_jobs)
        # special case of scipy API for tree.query
        if self.n_jobs is None:
            self.n_jobs = -1

    def _instanciate_vars(self, adjacency: Union[sparse.csr_matrix, np.ndarray], seeds: Union[np.ndarray, dict]):
        n = adjacency.shape[0]
        labels = check_seeds(seeds, n).astype(int)
        index_seed = np.argwhere(labels >= 0).ravel()
        index_remain = np.argwhere(labels < 0).ravel()
        labels_seed = labels[index_seed]

        embedding = self.embedding_method.fit_transform(adjacency)

        return index_seed, index_remain, labels_seed, embedding

    def _fit_core(self, n, labels_seed, embedding, index_seed, index_remain):
        n_seeds = len(labels_seed)
        embedding_seed = embedding[index_seed]
        embedding_remain = embedding[index_remain]
        n_neighbors = check_n_neighbors(self.n_neighbors, n_seeds)
        tree = cKDTree(embedding_seed, self.leaf_size)
        distances, neighbors = tree.query(embedding_remain, n_neighbors, self.tol_nn, self.p, n_jobs=self.n_jobs)

        if n_neighbors == 1:
            distances = distances[:, np.newaxis]
            neighbors = neighbors[:, np.newaxis]

        labels_neighbor = labels_seed[neighbors]
        index = (np.min(distances, axis=1) == 0)
        weights_neighbor = np.zeros_like(distances).astype(float)
        # take all seeds at distance zero, if any
        weights_neighbor[index] = (distances[index] == 0).astype(float)
        # assign weights with respect to distances for other
        weights_neighbor[~index] = 1 / np.power(distances[~index], self.factor_distance)

        # form the corresponding matrix
        row = list(np.repeat(index_remain, n_neighbors))
        col = list(labels_neighbor.ravel())
        data = list(weights_neighbor.ravel())

        row += list(index_seed)
        col += list(labels_seed)
        data += list(np.ones_like(index_seed))

        membership = normalize(sparse.csr_matrix((data, (row, col)), shape=(n, np.max(labels_seed) + 1)))
        membership_dense = membership.toarray()
        labels = np.argmax(membership_dense, axis=1)

        return membership, labels

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray], seeds: Union[np.ndarray, dict]) -> 'KNN':
        """Node classification by k-nearest neighbors in the embedding space.

        Parameters
        ----------
        adjacency :
            Adjacency or biadjacency matrix of the graph.
        seeds :
            Seed nodes. Can be a dict {node: label} or an array where "-1" means no label.

        Returns
        -------
        self: :class:`KNN`
        """
        n = adjacency.shape[0]
        index_seed, index_remain, labels_seed, embedding = self._instanciate_vars(adjacency, seeds)

        membership, labels = self._fit_core(n, labels_seed, embedding, index_seed, index_remain)

        self.membership_ = membership
        self.labels_ = labels

        return self


class BiKNN(KNN, BaseBiClassifier):
    """Node classification by K-nearest neighbors in the embedding space.

    * Bigraphs

    Parameters
    ----------
    embedding_method :
        Which algorithm to use to project the nodes in vector space. Default is ``GSVD``.
    n_neighbors :
        Number of nearest neighbors to consider.
    factor_distance :
        Power weighting factor :math:``\\alpha`` applied to the distance to each neighbor.
        Neighbor at distance :math:``d`` has weight :math:``1 / d^\\alpha``. Default is 2.
    leaf_size :
        Leaf size passed to KDTree.
    p :
        Which Minkowski p-norm to use. Default is 2 (Euclidean distance).
    tol_nn :
        Tolerance in nearest neighbors search; the k-th returned value is guaranteed to be no further
        than ``1 + tol_nn`` times the distance to the actual k-th nearest neighbor.
    n_jobs :
        Number of jobs to schedule for parallel processing. If -1 is given all processors are used.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each row.
    labels_row_ : np.ndarray
        Label of each row (copy of **labels_**).
    labels_col_ : np.ndarray
        Label of each column.
    membership_ : sparse.csr_matrix
        Membership matrix of rows.
    membership_row_ : sparse.csr_matrix
        Membership matrix of rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of columns.

    Example
    -------
    >>> from sknetwork.classification import BiKNN
    >>> from sknetwork.data import movie_actor
    >>> biknn = BiKNN(n_neighbors=2)
    >>> graph = movie_actor(metadata=True)
    >>> biadjacency = graph.biadjacency
    >>> seeds_row = {0: 0, 1: 2, 2: 1}
    >>> len(biknn.fit_transform(biadjacency, seeds_row))
    15
    >>> len(biknn.labels_col_)
    16
    """
    def __init__(self, embedding_method: BaseEmbedding = GSVD(10), n_neighbors: int = 5,
                 factor_distance: float = 2, leaf_size: int = 16, p: float = 2, tol_nn: float = 0.01, n_jobs: int = 1):
        super(BiKNN, self).__init__(embedding_method, n_neighbors, factor_distance, leaf_size, p, tol_nn, n_jobs)

    def _instanciate_vars(self, biadjacency: Union[sparse.csr_matrix, np.ndarray], seeds_row: Union[np.ndarray, dict],
                          seeds_col: Optional[Union[np.ndarray, dict]] = None):
        n_row, n_col = biadjacency.shape
        labels = stack_seeds(n_row, n_col, seeds_row, seeds_col).astype(int)
        index_seed = np.argwhere(labels >= 0).ravel()
        index_remain = np.argwhere(labels < 0).ravel()
        labels_seed = labels[index_seed]

        self.embedding_method.fit(biadjacency)
        embedding_row = self.embedding_method.embedding_row_
        embedding_col = self.embedding_method.embedding_col_
        embedding = np.vstack((embedding_row, embedding_col))

        return index_seed, index_remain, labels_seed, embedding

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray], seeds_row: Union[np.ndarray, dict],
            seeds_col: Optional[Union[np.ndarray, dict]] = None) -> 'BiKNN':
        """Node classification by k-nearest neighbors in the embedding space.

        Parameters
        ----------
        biadjacency :
            Biadjacency matrix of the graph.
        seeds_row :
            Seed rows. Can be a dict {node: label} or an array where "-1" means no label.
        seeds_col :
            Seed columns (optional). Same format.

        Returns
        -------
        self: :class:`BiKNN`
        """
        n_row, n_col = biadjacency.shape
        index_seed, index_remain, labels_seed, embedding = self._instanciate_vars(biadjacency, seeds_row, seeds_col)

        membership, labels = self._fit_core(n_row + n_col, labels_seed, embedding, index_seed, index_remain)
        self.membership_ = membership
        self.labels_ = labels
        self._split_vars(n_row)

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov, 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from abc import ABC

import numpy as np

from sknetwork.utils.base import Algorithm


class BaseRanking(Algorithm, ABC):
    """Base class for ranking algorithms.

    Attributes
    ----------
    scores_ : np.ndarray
        PageRank score of each node.
    """
    def __init__(self):
        self.scores_ = None

    def fit_transform(self, *args, **kwargs) -> np.ndarray:
        """Fit algorithm to data and return the scores. Same parameters as the ``fit`` method.

        Returns
        -------
        scores : np.ndarray
            Scores.
        """
        self.fit(*args, **kwargs)
        return self.scores_


class BaseBiRanking(BaseRanking, ABC):
    """Base class for ranking algorithms.

    Attributes
    ----------
    scores_ : np.ndarray
        PageRank score of each node.
    scores_row_ : np.ndarray
        PageRank score of each row (copy of **scores_**).
    scores_col_ : np.ndarray
        PageRank score of each column.
    """
    def __init__(self):
        super(BaseBiRanking, self).__init__()
        self.scores_row_ = None
        self.scores_col_ = None

    def _split_vars(self, n_row):
        self.scores_row_ = self.scores_[:n_row]
        self.scores_col_ = self.scores_[n_row:]
        self.scores_ = self.scores_row_



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 8, 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse

from sknetwork.linalg.sparse_lowrank import SparseLR


def check_csr_or_slr(adjacency):
    """Check if input is csr or SparseLR and raise an error otherwise."""
    if type(adjacency) not in [sparse.csr_matrix, SparseLR]:
        raise TypeError('Input must be a scipy CSR matrix or a SparseLR object.')
    else:
        return


def directed2undirected(adjacency: Union[sparse.csr_matrix, SparseLR],
                        weighted: bool = True) -> Union[sparse.csr_matrix, SparseLR]:
    """Adjacency matrix of the undirected graph associated with some directed graph.

    The new adjacency matrix becomes either:

    :math:`A+A^T` (default)

    or

    :math:`\\max(A,A^T)`

    If the initial adjacency matrix :math:`A` is binary, bidirectional edges have weight 2
    (first method, default) or 1 (second method).

    Parameters
    ----------
    adjacency :
        Adjacency matrix.
    weighted :
        If ``True``, return the sum of the weights in both directions of each edge.

    Returns
    -------
    new_adjacency :
        New adjacency matrix (same format as input).
    """
    check_csr_or_slr(adjacency)
    if type(adjacency) == sparse.csr_matrix:
        if weighted:
            new_adjacency = adjacency.astype(float)
            new_adjacency += adjacency.T
        else:
            new_adjacency = (adjacency + adjacency.T).astype(bool)
        new_adjacency.tocsr().sort_indices()
        return new_adjacency
    else:
        if weighted:
            new_tuples = [(y, x) for (x, y) in adjacency.low_rank_tuples]
            return SparseLR(directed2undirected(adjacency.sparse_mat), adjacency.low_rank_tuples + new_tuples)
        else:
            raise ValueError('This function only works with ``weighted=True`` for SparseLR objects.')


def bipartite2directed(biadjacency: Union[sparse.csr_matrix, SparseLR]) -> Union[sparse.csr_matrix, SparseLR]:
    """Adjacency matrix of the directed graph associated with a bipartite graph
    (with edges from one part to the other).

    The returned adjacency matrix is:

    :math:`A  = \\begin{bmatrix} 0 & B \\\\ 0 & 0 \\end{bmatrix}`

    where :math:`B` is the biadjacency matrix.

    Parameters
    ----------
    biadjacency :
        Biadjacency matrix of the graph.

    Returns
    -------
    adjacency :
        Adjacency matrix (same format as input).
    """
    check_csr_or_slr(biadjacency)
    n_row, n_col = biadjacency.shape
    if type(biadjacency) == sparse.csr_matrix:
        adjacency = sparse.bmat([[None, biadjacency], [sparse.csr_matrix((n_col, n_row)), None]], format='csr')
        adjacency.sort_indices()
        return adjacency
    else:
        new_tuples = [(np.hstack((x, np.zeros(n_col))), np.hstack((np.zeros(n_row), y)))
                      for (x, y) in biadjacency.low_rank_tuples]
        return SparseLR(bipartite2directed(biadjacency.sparse_mat), new_tuples)


def bipartite2undirected(biadjacency: Union[sparse.csr_matrix, SparseLR]) -> Union[sparse.csr_matrix, SparseLR]:
    """Adjacency matrix of a bigraph defined by its biadjacency matrix.

    The returned adjacency matrix is:

    :math:`A  = \\begin{bmatrix} 0 & B \\\\ B^T & 0 \\end{bmatrix}`

    where :math:`B` is the biadjacency matrix of the bipartite graph.

    Parameters
    ----------
    biadjacency:
        Biadjacency matrix of the graph.

    Returns
    -------
    adjacency :
        Adjacency matrix (same format as input).
    """
    check_csr_or_slr(biadjacency)
    if type(biadjacency) == sparse.csr_matrix:
        adjacency = sparse.bmat([[None, biadjacency], [biadjacency.T, None]], format='csr')
        adjacency.sort_indices()
        return adjacency
    else:
        n_row, n_col = biadjacency.shape
        new_tuples = []
        for (x, y) in biadjacency.low_rank_tuples:
            new_tuples.append((np.hstack((x, np.zeros(n_col))), np.hstack((np.zeros(n_row), y))))
            new_tuples.append((np.hstack((np.zeros(n_row), y)), np.hstack((x, np.zeros(n_col)))))
        return SparseLR(bipartite2undirected(biadjacency.sparse_mat), new_tuples)



"""classification module"""
from sknetwork.classification.base import BaseClassifier, BaseBiClassifier
from sknetwork.classification.diffusion import DiffusionClassifier, BiDiffusionClassifier, DirichletClassifier, \
    BiDirichletClassifier
from sknetwork.classification.knn import KNN, BiKNN
from sknetwork.classification.metrics import accuracy_score
from sknetwork.classification.pagerank import BiPageRankClassifier, PageRankClassifier
from sknetwork.classification.propagation import BiPropagation, Propagation



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May 31 2019
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <bonald@enst.fr>
"""
from typing import Union, Optional

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import LinearOperator

from sknetwork.linalg.ppr_solver import get_pagerank
from sknetwork.ranking.base import BaseRanking, BaseBiRanking
from sknetwork.utils.check import check_format, check_square, check_damping_factor
from sknetwork.utils.format import bipartite2undirected
from sknetwork.utils.seeds import seeds2probs, stack_seeds
from sknetwork.utils.verbose import VerboseMixin


class PageRank(BaseRanking, VerboseMixin):
    """PageRank of each node, corresponding to its frequency of visit by a random walk.

    The random walk restarts with some fixed probability. The restart distribution can be personalized by the user.
    This variant is known as Personalized PageRank.

    * Graphs
    * Digraphs

    Parameters
    ----------
    damping_factor : float
        Probability to continue the random walk.
    solver : str
        * ``'piteration'``, use power iteration for a given number of iterations.
        * ``'diteration'``, use asynchronous parallel diffusion for a given number of iterations.
        * ``'lanczos'``, use eigensolver with a given tolerance.
        * ``'bicgstab'``, use Biconjugate Gradient Stabilized method for a given tolerance.
        * ``'RH'``, use a Ruffini-Horner polynomial evaluation.
    n_iter : int
        Number of iterations for some solvers.
    tol : float
        Tolerance for the convergence of some solvers.

    Attributes
    ----------
    scores_ : np.ndarray
        PageRank score of each node.

    Example
    -------
    >>> from sknetwork.ranking import PageRank
    >>> from sknetwork.data import house
    >>> pagerank = PageRank()
    >>> adjacency = house()
    >>> seeds = {0: 1}
    >>> scores = pagerank.fit_transform(adjacency, seeds)
    >>> np.round(scores, 2)
    array([0.29, 0.24, 0.12, 0.12, 0.24])

    References
    ----------
    Page, L., Brin, S., Motwani, R., & Winograd, T. (1999). The PageRank citation ranking: Bringing order to the web.
    Stanford InfoLab.
    """
    def __init__(self, damping_factor: float = 0.85, solver: str = 'piteration', n_iter: int = 10, tol: float = 1e-6):
        super(PageRank, self).__init__()
        check_damping_factor(damping_factor)
        self.damping_factor = damping_factor
        self.solver = solver
        self.n_iter = n_iter
        self.tol = tol

    # noinspection PyTypeChecker
    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray, LinearOperator],
            seeds: Optional[Union[dict, np.ndarray]] = None) -> 'PageRank':
        """Fit algorithm to data.

        Parameters
        ----------
        adjacency :
            Adjacency matrix.
        seeds :
            Parameter to be used for Personalized PageRank.
            Restart distribution as a vector or a dict (node: weight).
            If ``None``, the uniform distribution is used (no personalization, default).

        Returns
        -------
        self: :class:`PageRank`
        """
        if not isinstance(adjacency, LinearOperator):
            adjacency = check_format(adjacency)
        check_square(adjacency)
        seeds = seeds2probs(adjacency.shape[0], seeds)
        self.scores_ = get_pagerank(adjacency, seeds, damping_factor=self.damping_factor, n_iter=self.n_iter,
                                    solver=self.solver, tol=self.tol)

        return self


class BiPageRank(PageRank, BaseBiRanking):
    """Compute the PageRank of each node through a random walk in the bipartite graph.

    * Bigraphs

    Parameters
    ----------
    damping_factor : float
        Probability to continue the random walk.
    solver : str
        * `piteration`, use power iteration for a given number of iterations.
        * `diteration`, use asynchronous parallel diffusion for a given number of iterations.
        * `lanczos`, use eigensolver for a given tolerance.
        * `bicgstab`, use Biconjugate Gradient Stabilized method for a given tolerance.
    n_iter : int
        Number of iterations for some solvers.
    tol : float
        Tolerance for the convergence of some solvers.

    Attributes
    ----------
    scores_ : np.ndarray
        PageRank score of each row.
    scores_row_ : np.ndarray
        PageRank score of each row (copy of **scores_**).
    scores_col_ : np.ndarray
        PageRank score of each column.

    Example
    -------
    >>> from sknetwork.ranking import BiPageRank
    >>> from sknetwork.data import star_wars
    >>> bipagerank = BiPageRank()
    >>> biadjacency = star_wars()
    >>> seeds = {0: 1}
    >>> scores = bipagerank.fit_transform(biadjacency, seeds)
    >>> np.round(scores, 2)
    array([0.45, 0.11, 0.28, 0.17])
    """
    def __init__(self, damping_factor: float = 0.85, solver: str = 'piteration', n_iter: int = 10, tol: float = 0):
        super(BiPageRank, self).__init__(damping_factor, solver, n_iter, tol)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray],
            seeds_row: Optional[Union[dict, np.ndarray]] = None, seeds_col: Optional[Union[dict, np.ndarray]] = None) \
            -> 'BiPageRank':
        """Fit algorithm to data.

        Parameters
        ----------
        biadjacency :
            Biadjacency matrix.
        seeds_row, seeds_col :
            Parameter to be used for Personalized BiPageRank.
            Restart distribution as vectors or dicts on rows, columns (node: weight).
            If both seeds_row and seeds_col are ``None`` (default), the uniform distribution on rows is used.

        Returns
        -------
        self: :class:`BiPageRank`
        """
        biadjacency = check_format(biadjacency)
        n_row, n_col = biadjacency.shape
        adjacency = bipartite2undirected(biadjacency)
        seeds = stack_seeds(n_row, n_col, seeds_row, seeds_col)

        PageRank.fit(self, adjacency, seeds)
        self._split_vars(n_row)

        self.scores_row_ /= self.scores_row_.sum()
        self.scores_col_ /= self.scores_col_.sum()
        self.scores_ = self.scores_row_

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
import numpy as np
from scipy.cluster.hierarchy import ward

from sknetwork.utils.base import Algorithm


class WardDense(Algorithm):
    """Hierarchical clustering by the Ward method based on SciPy.

    Attributes
    ----------
    dendrogram_ : np.ndarray (n - 1, 4)
        Dendrogram.

    References
    ----------
    * Ward, J. H., Jr. (1963). Hierarchical grouping to optimize an objective function.
      Journal of the American Statistical Association, 58, 236–244.

    * Murtagh, F., & Contreras, P. (2012). Algorithms for hierarchical clustering: an overview.
      Wiley Interdisciplinary Reviews: Data Mining and Knowledge Discovery, 2(1), 86-97.
    """
    def __init__(self):
        self.dendrogram_ = None

    def fit(self, x: np.ndarray) -> 'WardDense':
        """Apply algorithm to a dense matrix.

        Parameters
        ----------
        x:
            Data to cluster.

        Returns
        -------
        self: :class:`WardDense`
        """
        self.dendrogram_ = ward(x)
        return self

    def fit_transform(self, x: np.ndarray) -> np.ndarray:
        """Apply algorithm to a dense matrix and return the dendrogram.

        Parameters
        ----------
        x:
            Data to cluster.

        Returns
        -------
        dendrogram: np.ndarray
        """
        self.fit(x)
        return self.dendrogram_
